import os
from collections.abc import Collection
from pathlib import Path
from typing import Iterable, Optional, Self

from ai.backend.common.docker import KernelFeatures
from ai.backend.common.types import KernelId


class ScratchUtil:
    @classmethod
    def scratch_dir(cls, scratch_root: Path, kernel_id: KernelId) -> Path:
        return (scratch_root / str(kernel_id)).resolve()

    @classmethod
    def scratch_file(cls, scratch_root: Path, kernel_id: KernelId) -> Path:
        return (cls.scratch_dir(scratch_root, kernel_id) / f"{kernel_id}.img").resolve()

    @classmethod
    def tmp_dir(cls, scratch_root: Path, kernel_id: KernelId) -> Path:
        return (scratch_root / f"{kernel_id}_tmp").resolve()

    @classmethod
    def work_dir(cls, scratch_root: Path, kernel_id: KernelId) -> Path:
        return (cls.scratch_dir(scratch_root, kernel_id) / "work").resolve()

    @classmethod
    def config_dir(cls, scratch_root: Path, kernel_id: KernelId) -> Path:
        return (cls.scratch_dir(scratch_root, kernel_id) / "config").resolve()


class PathOwnerDeterminer:
    """
    Determines the final UID and GID for a path based on overrides and kernel features.
    """

    def __init__(self, kernel_uid: int, kernel_gid: int, do_uid_match: bool) -> None:
        self._do_uid_match = do_uid_match
        self._kernel_uid = kernel_uid
        self._kernel_gid = kernel_gid

    @classmethod
    def by_kernel_features(
        cls,
        kernel_uid: int,
        kernel_gid: int,
        kernel_features: Collection[str],
    ) -> Self:
        """
        Create a PathOwnerDeterminer based on kernel features.
        If UID_MATCH is enabled, it uses kernel_uid/gid.
        Otherwise, it uses the current file's UID/GID.
        """
        do_uid_match = KernelFeatures.UID_MATCH in kernel_features
        return cls(kernel_uid=kernel_uid, kernel_gid=kernel_gid, do_uid_match=do_uid_match)

    def determine(
        self, path: Path, *, uid_override: Optional[int], gid_override: Optional[int]
    ) -> tuple[int, int]:
        """
        Determine the final UID and GID for a path.
        If uid_override/gid_override are provided, they are used.
        If not, and if uid-match feature is enabled, it uses kernel_uid/gid.
        Otherwise, it uses the current file's UID/GID.
        """
        pstat = path.stat()
        final_uid = (
            uid_override
            if uid_override is not None
            else (self._kernel_uid if self._do_uid_match else pstat.st_uid)
        )
        final_gid = (
            gid_override
            if gid_override is not None
            else (self._kernel_gid if self._do_uid_match else pstat.st_gid)
        )
        return final_uid, final_gid


class ChownUtil:
    def __init__(self) -> None:
        self._do_chown = 0 == os.geteuid()

    def chown_path(self, path: Path, uid: int, gid: int) -> None:
        """
        Change ownership of a single path if running as root.
        """
        if self._do_chown:
            os.chown(path, uid, gid)

    def chown_paths(self, paths: Iterable[Path], uid: int, gid: int) -> None:
        """
        Change ownership of multiple paths if running as root.
        """
        if self._do_chown:
            for p in paths:
                self.chown_path(p, uid, gid)



from __future__ import annotations

import enum
import fnmatch
import itertools
import logging
import operator
from abc import ABCMeta, abstractmethod
from collections import defaultdict
from decimal import ROUND_DOWN, Decimal
from typing import (
    TYPE_CHECKING,
    FrozenSet,
    Iterable,
    Mapping,
    MutableMapping,
    Optional,
    Sequence,
    TypeVar,
    final,
)

import attr
import more_itertools

from ai.backend.common.exception import ConfigurationError
from ai.backend.common.types import DeviceId, DeviceName, SlotName, SlotTypes
from ai.backend.logging import BraceStyleAdapter

from .affinity_map import AffinityHint, AffinityPolicy
from .exception import (
    FractionalResourceFragmented,
    InsufficientResource,
    InvalidResourceArgument,
    InvalidResourceCombination,
    NotMultipleOfQuantum,
    ResourceError,
)

if TYPE_CHECKING:
    from .resources import AbstractComputeDevice

log = BraceStyleAdapter(logging.getLogger(__spec__.name))
log_alloc_map: bool = False
T = TypeVar("T")


class AllocationStrategy(enum.Enum):
    FILL = 0
    EVENLY = 1


@attr.define()
class DeviceSlotInfo:
    slot_type: SlotTypes
    slot_name: SlotName
    amount: Decimal


def distribute(num_items: int, groups: Sequence[T]) -> Mapping[T, int]:
    base, extra = divmod(num_items, len(groups))
    return dict(
        zip(
            groups,
            ((base + (1 if i < extra else 0)) for i in range(len(groups))),
        )
    )


def round_down(from_dec: Decimal, with_dec: Decimal):
    remainder = from_dec.remainder_near(with_dec)
    if remainder < 0:
        remainder += with_dec
    return from_dec - remainder


class AbstractAllocMap(metaclass=ABCMeta):
    device_slots: Mapping[DeviceId, DeviceSlotInfo]
    device_mask: FrozenSet[DeviceId]
    exclusive_slot_types: Iterable[SlotName]
    allocations: MutableMapping[SlotName, MutableMapping[DeviceId, Decimal]]

    def __init__(
        self,
        *,
        device_slots: Optional[Mapping[DeviceId, DeviceSlotInfo]] = None,
        device_mask: Optional[Iterable[DeviceId]] = None,
        exclusive_slot_types: Optional[Iterable[SlotName]] = None,
    ) -> None:
        self.exclusive_slot_types = exclusive_slot_types or {}
        self.device_slots = device_slots or {}
        self.slot_types = {info.slot_name: info.slot_type for info in self.device_slots.values()}
        self.device_mask = frozenset(device_mask) if device_mask is not None else frozenset()
        self.allocations = defaultdict(lambda: defaultdict(Decimal))
        for dev_id, dev_slot_info in self.device_slots.items():
            self.allocations[dev_slot_info.slot_name][dev_id] = Decimal(0)

    @final
    def clear(self) -> None:
        self.allocations.clear()
        for dev_id, dev_slot_info in self.device_slots.items():
            self.allocations[dev_slot_info.slot_name][dev_id] = Decimal(0)

    @final
    def check_exclusive(self, a: SlotName, b: SlotName) -> bool:
        if not self.exclusive_slot_types:
            return False
        if a == b:
            return False
        a_in_exclusive_set = a in self.exclusive_slot_types
        b_in_exclusive_set = b in self.exclusive_slot_types
        if a_in_exclusive_set and b_in_exclusive_set:
            # fast-path for exact match
            return True
        for t in self.exclusive_slot_types:
            if "*" in t:
                a_in_exclusive_set = a_in_exclusive_set or fnmatch.fnmatchcase(a, t)
                b_in_exclusive_set = b_in_exclusive_set or fnmatch.fnmatchcase(b, t)
        return a_in_exclusive_set and b_in_exclusive_set

    def format_current_allocations(self) -> str:
        bufs = []
        for slot_name, per_device_alloc in self.allocations.items():
            bufs.append(f"slot[{slot_name}]:")
            for device_id, alloc in per_device_alloc.items():
                bufs.append(f"  {device_id}: {alloc}")
        return "\n".join(bufs)

    @final
    def get_current_allocations(
        self, affinity_hint: Optional[AffinityHint], slot_name: SlotName
    ) -> Sequence[tuple[DeviceId, Decimal]]:
        device_name = DeviceName(slot_name.partition(".")[0])

        if affinity_hint is None:  # for legacy
            return sorted(
                self.allocations[slot_name].items(),  # k: slot_name, v: per-device alloc
                key=lambda pair: self.device_slots[pair[0]].amount - pair[1],
                reverse=True,
            )

        # Use the affinity hint to reorder the device sets to prioritize allocation.
        if not affinity_hint.devices:
            # In the first resource slot during allocation,
            # we extract the device clusters based on the logical distances between each pair.
            primary_sets = affinity_hint.affinity_map.get_device_clusters_with_lowest_distance(
                device_name
            )
            secondary_set: Sequence[AbstractComputeDevice] = []
        else:
            # In the subsequent resource slots during allocation,
            # we build the device clusters based on the device sets allocated for the previous resource slot.
            # This logic is particularly important to seamlessly support specific GDS solutions like WEKA,
            # as it requires alignment of CPU cores and GPU devices to have consistent NUMA nodes.
            primary_sets, secondary_set = affinity_hint.affinity_map.get_distance_ordered_neighbors(
                affinity_hint.devices, device_name
            )

        if not primary_sets:
            # In normal conditions, we should have at least one primary set.
            # The errorneous case may happen when:
            # - the device_name is not explicitly set, and
            # - auto-generated device_name differs from the reported resource slot name.
            #   (e.g., "NPUPlusDevice" -> "npuplus" while the slot name is "npu-plus.device")
            # Here, we report this failure explicitly so that we could check the
            # potentially missing explicit device_name in the accelerator plugins.
            detected_device_names = {
                device.device_name for device in affinity_hint.affinity_map.nodes.keys()
            }
            raise ConfigurationError({
                "AbstractAllocMap.get_current_allocation()": (
                    f"No suitable devices found "
                    f"(searched: {device_name}, found: {', '.join(detected_device_names)})"
                )
            })

        # Let it prefer the largest primary set based on the remaining capacity per NUMA node.
        # Note that each primary device set is already included within the same numa node.
        primary_sets = sorted(
            primary_sets,
            key=lambda device_set: sum(
                (self.allocations[slot_name][device.device_id] for device in device_set),
                start=Decimal(0),
            ),
        )

        def convert_to_sorted_dev_alloc(
            device_set: Iterable[AbstractComputeDevice],
        ) -> list[tuple[DeviceId, Decimal]]:
            device_ids = {d.device_id for d in device_set}
            return sorted(
                (
                    (device_id, alloc)
                    for device_id, alloc in self.allocations[slot_name].items()
                    if device_id in device_ids
                ),
                key=lambda pair: self.device_slots[pair[0]].amount - pair[1],
                reverse=True,
            )

        # Inside each device cluster, let it prefer the devices with most remaining capacity.
        primary_sorted_dev_allocs = [
            convert_to_sorted_dev_alloc(primary_set) for primary_set in primary_sets
        ]
        secondary_sorted_dev_alloc = convert_to_sorted_dev_alloc(secondary_set)

        if not affinity_hint.devices:  # first-allocated device
            match affinity_hint.policy:
                case AffinityPolicy.PREFER_SINGLE_NODE:
                    return [
                        *itertools.chain(*primary_sorted_dev_allocs)
                    ] + secondary_sorted_dev_alloc
                case AffinityPolicy.INTERLEAVED:
                    return [
                        *more_itertools.interleave_longest(*primary_sorted_dev_allocs)
                    ] + secondary_sorted_dev_alloc
        else:
            return [
                *more_itertools.interleave_longest(*primary_sorted_dev_allocs)
            ] + secondary_sorted_dev_alloc

    @final
    def update_affinity_hint(
        self,
        device_alloc: Mapping[DeviceId, Decimal],
        affinity_hint: Optional[AffinityHint] = None,
    ) -> None:
        if affinity_hint is None:
            return
        hint_for_next_allocation: list[AbstractComputeDevice] = []
        for dev_id, alloc in device_alloc.items():
            if alloc == Decimal(0):
                continue
            for dev in affinity_hint.affinity_map.nodes:
                if dev.device_id == dev_id:
                    hint_for_next_allocation.append(dev)
        affinity_hint.devices = hint_for_next_allocation

    @abstractmethod
    def allocate(
        self,
        slots: Mapping[SlotName, Decimal],
        *,
        affinity_hint: Optional[AffinityHint] = None,
        context_tag: Optional[str] = None,
    ) -> Mapping[SlotName, Mapping[DeviceId, Decimal]]:
        """
        Allocate the given amount of resources.

        For a slot type, there may be multiple different devices which can allocate resources
        in the given slot type.  An implementation of alloc map finds suitable match from the
        remaining capacities of those devices.

        Returns a mapping from each requested slot to the allocations per device.
        """
        pass

    @abstractmethod
    def apply_allocation(
        self,
        existing_alloc: Mapping[SlotName, Mapping[DeviceId, Decimal]],
    ) -> None:
        """
        Apply the given allocation restored from disk or other persistent storages.
        """
        pass

    @abstractmethod
    def free(
        self,
        existing_alloc: Mapping[SlotName, Mapping[DeviceId, Decimal]],
    ) -> None:
        """
        Free the allocated resources using the token returned when the allocation
        occurred.
        """
        pass


class DiscretePropertyAllocMap(AbstractAllocMap):
    """
    An allocation map using discrete property.
    The user must pass a "property function" which returns a desired resource
    property from the device object.

    e.g., 1.0 means 1 device, 2.0 means 2 devices, etc.
    (no fractions allowed)
    """

    def __init__(
        self,
        *args,
        allocation_strategy: AllocationStrategy = AllocationStrategy.EVENLY,
        **kwargs,
    ) -> None:
        self.allocation_strategy = allocation_strategy
        self._allocate_impl = {
            AllocationStrategy.FILL: self._allocate_by_filling,
            AllocationStrategy.EVENLY: self._allocate_evenly,
        }
        super().__init__(*args, **kwargs)

    def allocate(
        self,
        slots: Mapping[SlotName, Decimal],
        *,
        affinity_hint: Optional[AffinityHint] = None,
        context_tag: Optional[str] = None,
    ) -> Mapping[SlotName, Mapping[DeviceId, Decimal]]:
        # prune zero alloc slots
        requested_slots = {k: v for k, v in slots.items() if v > 0}

        # check exclusive
        for slot_name_a in requested_slots.keys():
            for slot_name_b in requested_slots.keys():
                if self.check_exclusive(slot_name_a, slot_name_b):
                    raise InvalidResourceCombination(
                        f"Slots {slot_name_a} and {slot_name_b} cannot be allocated at the same"
                        " time."
                    )

        # check unique
        for slot_name, alloc in requested_slots.items():
            slot_type = self.slot_types.get(slot_name, SlotTypes.COUNT)
            if slot_type in (SlotTypes.COUNT, SlotTypes.BYTES):
                pass
            elif slot_type == SlotTypes.UNIQUE:
                if alloc != Decimal(1):
                    raise InvalidResourceArgument(
                        f"You may allocate only 1 for the unique-type slot {slot_name}",
                    )

        return self._allocate_impl[self.allocation_strategy](
            requested_slots,
            affinity_hint=affinity_hint,
            context_tag=context_tag,
        )

    def _allocate_by_filling(
        self,
        requested_slots: Mapping[SlotName, Decimal],
        *,
        affinity_hint: Optional[AffinityHint] = None,
        context_tag: Optional[str] = None,
    ) -> Mapping[SlotName, Mapping[DeviceId, Decimal]]:
        allocation: dict[SlotName, dict[DeviceId, Decimal]] = {}
        for slot_name, requested_alloc in requested_slots.items():
            slot_allocation: dict[DeviceId, Decimal] = {}
            sorted_dev_allocs = self.get_current_allocations(affinity_hint, slot_name)
            if log_alloc_map:
                log.debug(
                    "DiscretePropertyAllocMap(FILL): allocating {} {}", slot_name, requested_alloc
                )
                log.debug("DiscretePropertyAllocMap(FILL): current-alloc: {!r}", sorted_dev_allocs)

            total_allocatable = int(0)
            remaining_alloc = Decimal(requested_alloc).normalize()

            # fill up starting from the most free devices considering affinity hint
            for dev_id, current_alloc in sorted_dev_allocs:
                current_alloc = self.allocations[slot_name][dev_id]
                assert slot_name == self.device_slots[dev_id].slot_name
                total_allocatable += int(self.device_slots[dev_id].amount - current_alloc)
            if total_allocatable < requested_alloc:
                raise InsufficientResource(
                    "DiscretePropertyAllocMap: insufficient allocatable amount!",
                    context_tag=context_tag,
                    slot_name=slot_name,
                    requested_alloc=requested_alloc,
                    total_allocatable=total_allocatable,
                    allocation=allocation,
                )
            for dev_id, current_alloc in sorted_dev_allocs:
                current_alloc = self.allocations[slot_name][dev_id]
                allocatable = self.device_slots[dev_id].amount - current_alloc
                if allocatable > 0:
                    allocated = Decimal(min(remaining_alloc, allocatable))
                    slot_allocation[dev_id] = allocated
                    self.allocations[slot_name][dev_id] += allocated
                    remaining_alloc -= allocated
                if remaining_alloc == 0:
                    break
            allocation[slot_name] = slot_allocation
            self.update_affinity_hint(slot_allocation, affinity_hint)

        return allocation

    def _allocate_evenly(
        self,
        requested_slots: Mapping[SlotName, Decimal],
        *,
        affinity_hint: Optional[AffinityHint] = None,
        context_tag: Optional[str] = None,
    ) -> Mapping[SlotName, Mapping[DeviceId, Decimal]]:
        allocation: dict[SlotName, dict[DeviceId, Decimal]] = {}

        for slot_name, requested_alloc in requested_slots.items():
            new_alloc: MutableMapping[DeviceId, Decimal] = defaultdict(Decimal)
            remaining_alloc = int(Decimal(requested_alloc))
            if log_alloc_map:
                log.debug(
                    "DiscretePropertyAllocMap(EVENLY): allocating {} {}", slot_name, requested_alloc
                )

            repeats = 0
            while remaining_alloc > 0:
                # prevent infinite loop
                if repeats >= 100:
                    raise ResourceError("too many repeats until allocation")

                # sort the devices by the affinity hint
                sorted_dev_allocs = self.get_current_allocations(affinity_hint, slot_name)
                if log_alloc_map and repeats == 0:
                    log.debug(
                        "DiscretePropertyAllocMap(EVENLY): current-alloc: {!r}", sorted_dev_allocs
                    )

                # calculate remaining slots per device
                total_allocatable = int(
                    sum(
                        self.device_slots[dev_id].amount - current_alloc - new_alloc[dev_id]
                        for dev_id, current_alloc in self.allocations[slot_name].items()
                    )
                )
                # if the sum of remaining slot is less than the remaining alloc, fail.
                if total_allocatable < remaining_alloc:
                    raise InsufficientResource(
                        "DiscretePropertyAllocMap: insufficient allocatable amount!",
                        context_tag=context_tag,
                        slot_name=slot_name,
                        requested_alloc=requested_alloc,
                        total_allocatable=total_allocatable,
                        allocation=allocation,
                    )

                # calculate the amount to spread out
                nonzero_devs = [
                    dev_id
                    for dev_id, current_alloc in sorted_dev_allocs
                    if self.device_slots[dev_id].amount - current_alloc - new_alloc[dev_id] > 0
                ]
                if len(nonzero_devs) == 0:
                    raise InsufficientResource(
                        "DiscretePropertyAllocMap: insufficient allocatable candidate devices!",
                        context_tag=context_tag,
                        slot_name=slot_name,
                        requested_alloc=requested_alloc,
                        total_allocatable=total_allocatable,
                        allocation=allocation,
                    )
                initial_diffs = distribute(remaining_alloc, nonzero_devs)
                diffs = {
                    dev_id: min(
                        int(self.device_slots[dev_id].amount - current_alloc - new_alloc[dev_id]),
                        initial_diffs.get(dev_id, 0),
                    )
                    for dev_id, current_alloc in self.allocations[slot_name].items()
                }

                # distribute the remainig alloc to the remaining slots.
                for dev_id, current_alloc in sorted_dev_allocs:
                    diff = diffs[dev_id]
                    new_alloc[dev_id] += diff
                    remaining_alloc -= diff
                    if remaining_alloc == 0:
                        break

                repeats += 1

            for dev_id, allocated in new_alloc.items():
                self.allocations[slot_name][dev_id] += allocated
            allocation[slot_name] = {k: v for k, v in new_alloc.items() if v > 0}
            self.update_affinity_hint(new_alloc, affinity_hint)
            if log_alloc_map:
                log.debug("DiscretePropertyAllocMap(EVENLY): new-alloc: {!r}", new_alloc)

        return allocation

    def apply_allocation(
        self,
        existing_alloc: Mapping[SlotName, Mapping[DeviceId, Decimal]],
    ) -> None:
        for slot_name, per_device_alloc in existing_alloc.items():
            for device_id, alloc in per_device_alloc.items():
                self.allocations[slot_name][device_id] += alloc

    def free(
        self,
        existing_alloc: Mapping[SlotName, Mapping[DeviceId, Decimal]],
    ) -> None:
        for slot_name, per_device_alloc in existing_alloc.items():
            for device_id, alloc in per_device_alloc.items():
                self.allocations[slot_name][device_id] -= alloc


class FractionAllocMap(AbstractAllocMap):
    def __init__(
        self,
        *args,
        allocation_strategy: AllocationStrategy = AllocationStrategy.EVENLY,
        quantum_size: Decimal = Decimal("0.01"),
        enforce_physical_continuity: bool = True,
        **kwargs,
    ) -> None:
        self.allocation_strategy = allocation_strategy
        self.quantum_size = quantum_size
        self.enforce_physical_continuity = enforce_physical_continuity
        self._allocate_impl = {
            AllocationStrategy.FILL: self._allocate_by_filling,
            AllocationStrategy.EVENLY: self._allocate_evenly,
        }
        super().__init__(*args, **kwargs)
        self.digits = Decimal(10) ** -2  # decimal points that is supported by agent
        self.powers = Decimal(100)  # reciprocal of self.digits

    def allocate(
        self,
        slots: Mapping[SlotName, Decimal],
        *,
        affinity_hint: Optional[AffinityHint] = None,
        context_tag: Optional[str] = None,
        min_memory: Decimal = Decimal("0.01"),
        allow_resource_fragmentation: bool = True,
    ) -> Mapping[SlotName, Mapping[DeviceId, Decimal]]:
        # prune zero alloc slots
        requested_slots = {k: v for k, v in slots.items() if v > 0}

        # check exclusive
        for slot_name_a in requested_slots.keys():
            for slot_name_b in requested_slots.keys():
                if self.check_exclusive(slot_name_a, slot_name_b):
                    raise InvalidResourceCombination(
                        f"Slots {slot_name_a} and {slot_name_b} cannot be allocated at the same"
                        " time.",
                    )

        calculated_alloc_map = self._allocate_impl[self.allocation_strategy](
            requested_slots,
            affinity_hint=affinity_hint,
            context_tag=context_tag,
            min_memory=min_memory,
            allow_slot_fragmentation=allow_resource_fragmentation,
        )
        actual_alloc_map: dict[SlotName, dict[DeviceId, Decimal]] = {}
        for slot_name, alloc in calculated_alloc_map.items():
            actual_alloc: dict[DeviceId, Decimal] = {}
            for dev_id, val in alloc.items():
                self.allocations[slot_name][dev_id] = round_down(
                    self.allocations[slot_name][dev_id], self.quantum_size
                )
                actual_alloc[dev_id] = round_down(val, self.quantum_size)
            if sum(actual_alloc.values()) == 0 and requested_slots[slot_name] > 0:
                raise NotMultipleOfQuantum(
                    f"Requested resource amount for {slot_name} is"
                    f" {requested_slots[slot_name]} but actual calculated amount is zero. This"
                    " can happen if user requests resource amount smaller than target device's"
                    " quantum size.",
                )
            actual_alloc_map[slot_name] = actual_alloc

        return actual_alloc_map

    def ensure_slot_not_fragmented(
        self,
        slot_name: SlotName,
        sorted_dev_allocs: Sequence[tuple[DeviceId, Decimal]],
        alloc: Decimal,
        *,
        context_tag: Optional[str] = None,
    ) -> None:
        if len(sorted_dev_allocs) == 0:
            raise FractionalResourceFragmented(
                "FractionAllocMap: no device left for allocation!",
                context_tag=context_tag,
                slot_name=slot_name,
                requested_alloc=alloc,
                dev_allocs=sorted_dev_allocs,
            )
        most_free_dev_id, most_free_device_alloc = sorted_dev_allocs[0]
        most_free_device_allocatable = (
            self.device_slots[most_free_dev_id].amount - most_free_device_alloc
        )
        if most_free_device_allocatable < alloc:
            raise FractionalResourceFragmented(
                "FractionAllocMap: refusing to create kernel with fractional resource fragmented!",
                context_tag=context_tag,
                slot_name=slot_name,
                requested_alloc=alloc,
                dev_allocs=sorted_dev_allocs,
            )

    def _allocate_by_filling(
        self,
        requested_slots: Mapping[SlotName, Decimal],
        *,
        affinity_hint: Optional[AffinityHint] = None,
        context_tag: Optional[str] = None,
        min_memory: Decimal = Decimal(0.01),
        allow_slot_fragmentation: bool = True,
    ) -> Mapping[SlotName, Mapping[DeviceId, Decimal]]:
        allocation: dict[SlotName, dict[DeviceId, Decimal]] = {}
        for slot_name, alloc in requested_slots.items():
            slot_allocation: dict[DeviceId, Decimal] = {}

            # fill up starting from the most free devices
            sorted_dev_allocs = self.get_current_allocations(affinity_hint, slot_name)
            if not allow_slot_fragmentation:
                self.ensure_slot_not_fragmented(
                    slot_name,
                    sorted_dev_allocs,
                    alloc,
                    context_tag=context_tag,
                )

            if log_alloc_map:
                log.debug("FractionAllocMap(FILL): allocating {} {}", slot_name, alloc)
                log.debug("FractionAllocMap(FILL): current-alloc: {!r}", sorted_dev_allocs)

            slot_type = self.slot_types.get(slot_name, SlotTypes.COUNT)
            if slot_type in (SlotTypes.COUNT, SlotTypes.BYTES):
                pass
            elif slot_type == SlotTypes.UNIQUE:
                if alloc != Decimal(1):
                    raise InvalidResourceArgument(
                        f"You may allocate only 1 for the unique-type slot {slot_name}",
                    )
            total_allocatable = Decimal(0)
            remaining_alloc = Decimal(alloc).normalize()

            for dev_id, current_alloc in sorted_dev_allocs:
                current_alloc = self.allocations[slot_name][dev_id]
                assert slot_name == self.device_slots[dev_id].slot_name
                total_allocatable += self.device_slots[dev_id].amount - current_alloc
            if total_allocatable < alloc:
                raise InsufficientResource(
                    "FractionAllocMap: insufficient allocatable amount!",
                    context_tag=context_tag,
                    slot_name=slot_name,
                    requested_alloc=alloc,
                    total_allocatable=total_allocatable,
                    allocation=allocation,
                )
            for dev_id, current_alloc in sorted_dev_allocs:
                current_alloc = self.allocations[slot_name][dev_id]
                allocatable = self.device_slots[dev_id].amount - current_alloc
                if allocatable > 0:
                    allocated = min(remaining_alloc, allocatable)
                    slot_allocation[dev_id] = allocated
                    self.allocations[slot_name][dev_id] += allocated
                    remaining_alloc -= allocated
                if remaining_alloc <= 0:
                    break

            allocation[slot_name] = slot_allocation
            self.update_affinity_hint(slot_allocation, affinity_hint)
        return allocation

    def _allocate_evenly(
        self,
        requested_slots: Mapping[SlotName, Decimal],
        *,
        affinity_hint: Optional[AffinityHint] = None,
        context_tag: Optional[str] = None,
        min_memory: Decimal = Decimal(0.01),
        allow_slot_fragmentation: bool = True,
    ) -> Mapping[SlotName, Mapping[DeviceId, Decimal]]:
        # higher value means more even with 0 being the highest value
        def measure_evenness(
            alloc_map: Mapping[DeviceId, Decimal],
        ) -> Decimal:
            alloc_arr = sorted([alloc_map[dev_id] for dev_id in alloc_map])
            evenness_score = Decimal(0).quantize(self.digits)
            for idx in range(len(alloc_arr) - 1):
                evenness_score += abs(alloc_arr[idx + 1] - alloc_arr[idx])
            return -evenness_score

        # higher value means more fragmented
        # i.e. the number of unusable resources is higher
        def measure_fragmentation(
            allocation: Mapping[DeviceId, Decimal],
            min_memory: Decimal,
        ) -> int:
            fragmentation_arr = [
                self.device_slots[dev_id].amount - allocation[dev_id] for dev_id in allocation
            ]
            return sum(
                self.digits < v.quantize(self.digits) < min_memory.quantize(self.digits)
                for v in fragmentation_arr
            )

        # evenly distributes remaining_alloc across dev_allocs
        def distribute_evenly(
            dev_allocs: list[tuple[DeviceId, Decimal]],
            remaining_alloc: Decimal,
            allocation: dict[DeviceId, Decimal],
        ) -> None:
            n_devices = len(dev_allocs)
            for dev_id, _ in dev_allocs:
                dev_allocation = remaining_alloc / n_devices
                dev_allocation = dev_allocation.quantize(self.digits, rounding=ROUND_DOWN)
                allocation[dev_id] = dev_allocation

            # need to take care of decimals
            remainder = round(
                remaining_alloc * self.powers - dev_allocation * n_devices * self.powers
            )
            for idx in range(remainder):
                dev_id, _ = dev_allocs[idx]
                allocation[dev_id] += self.digits

        # allocates remaining_alloc across multiple devices i.e. dev_allocs
        # all devices in dev_allocs are being used
        def allocate_across_devices(
            dev_allocs: list[tuple[DeviceId, Decimal]],
            remaining_alloc: Decimal,
            slot_name: str,
        ) -> dict[DeviceId, Decimal]:
            slot_allocation: dict[DeviceId, Decimal] = {}
            n_devices = len(dev_allocs)
            idx = n_devices - 1  # check from the device with smallest allocatable resource
            while n_devices > 0:
                dev_id, current_alloc = dev_allocs[idx]
                allocatable = self.device_slots[dev_id].amount - current_alloc
                # if the remaining_alloc can be allocated to evenly among remaining devices
                if allocatable >= remaining_alloc / n_devices:
                    break
                slot_allocation[dev_id] = allocatable.quantize(self.digits)
                remaining_alloc -= allocatable
                idx -= 1
                n_devices -= 1

            if n_devices > 0:
                distribute_evenly(dev_allocs[:n_devices], remaining_alloc, slot_allocation)

            return slot_allocation

        min_memory = min_memory.quantize(self.digits)
        allocation: dict[SlotName, dict[DeviceId, Decimal]] = {}
        for slot_name, alloc in requested_slots.items():
            slot_allocation: dict[DeviceId, Decimal] = {}
            remaining_alloc = Decimal(alloc).normalize()
            sorted_dev_allocs = self.get_current_allocations(affinity_hint, slot_name)
            if not allow_slot_fragmentation:
                self.ensure_slot_not_fragmented(
                    slot_name,
                    sorted_dev_allocs,
                    alloc,
                    context_tag=context_tag,
                )

            # do not consider devices whose remaining resource under min_memory
            sorted_dev_allocs = list(
                filter(
                    lambda pair: self.device_slots[pair[0]].amount - pair[1] >= min_memory,
                    sorted_dev_allocs,
                )
            )

            if log_alloc_map:
                log.debug("FractionAllocMap(EVENLY): allocating {} {}", slot_name, alloc)
                log.debug("FractionAllocMap(EVENLY): current-alloc: {!r}", sorted_dev_allocs)

            # check if there is enough resource for allocation
            total_allocatable = Decimal(0)
            for dev_id, current_alloc in sorted_dev_allocs:
                current_alloc = self.allocations[slot_name][dev_id]
                total_allocatable += self.device_slots[dev_id].amount - current_alloc
            if total_allocatable.quantize(self.digits) < remaining_alloc.quantize(self.digits):
                raise InsufficientResource(
                    "FractionAllocMap: insufficient allocatable amount!",
                    context_tag=context_tag,
                    slot_name=slot_name,
                    requested_alloc=alloc,
                    total_allocatable=total_allocatable,
                    allocation=allocation,
                )

            # allocate resources
            if (
                remaining_alloc
                <= self.device_slots[sorted_dev_allocs[0][0]].amount - sorted_dev_allocs[0][1]
            ):
                # if remaining_alloc fits in one device
                for dev_id, current_alloc in sorted_dev_allocs[::-1]:
                    allocatable = self.device_slots[dev_id].amount - current_alloc
                    if remaining_alloc <= allocatable:
                        slot_allocation[dev_id] = remaining_alloc.quantize(self.digits)
                        break
            else:
                # need to distribute across devices
                # calculate the minimum number of required devices
                n_devices, allocated = 0, Decimal(0)
                for dev_id, current_alloc in sorted_dev_allocs:
                    n_devices += 1
                    allocated += self.device_slots[dev_id].amount - current_alloc
                    if allocated.quantize(self.digits) >= remaining_alloc.quantize(self.digits):
                        break
                # need to check from using minimum number of devices to using all devices
                # evenness must be non-decreasing with the increase of window size
                best_alloc_candidate_arr = []
                for n_dev in range(n_devices, len(sorted_dev_allocs) + 1):
                    allocatable = sum(
                        map(
                            lambda x: self.device_slots[x[0]].amount - x[1],
                            sorted_dev_allocs[:n_dev],
                        ),
                        start=Decimal(0),
                    )
                    # choose the best allocation from all possible allocation candidates
                    alloc_candidate = allocate_across_devices(
                        sorted_dev_allocs[:n_dev],
                        remaining_alloc,
                        slot_name,
                    )
                    max_evenness = measure_evenness(alloc_candidate)
                    # three criteria to decide allocation are
                    # eveness, number of resources used, and amount of fragmentatino
                    alloc_candidate_arr = [
                        (
                            alloc_candidate,
                            max_evenness,
                            -len(alloc_candidate),
                            -measure_fragmentation(alloc_candidate, min_memory),
                        ),
                    ]
                    for idx in range(1, len(sorted_dev_allocs) - n_dev + 1):
                        # update amount of allocatable space
                        allocatable -= (
                            self.device_slots[sorted_dev_allocs[idx - 1][0]].amount
                            - sorted_dev_allocs[idx - 1][1]
                        )
                        allocatable += (
                            self.device_slots[sorted_dev_allocs[idx + n_dev - 1][0]].amount
                            - sorted_dev_allocs[idx + n_dev - 1][1]
                        )
                        # break if not enough resource
                        if allocatable.quantize(self.digits) < remaining_alloc.quantize(
                            self.digits
                        ):
                            break
                        alloc_candidate = allocate_across_devices(
                            sorted_dev_allocs[idx : idx + n_dev],
                            remaining_alloc,
                            slot_name,
                        )
                        # evenness gets worse (or same at best) as the allocatable gets smaller
                        evenness_score = measure_evenness(alloc_candidate)
                        if evenness_score < max_evenness:
                            break
                        alloc_candidate_arr.append(
                            (
                                alloc_candidate,
                                evenness_score,
                                -len(alloc_candidate),
                                -measure_fragmentation(alloc_candidate, min_memory),
                            ),
                        )
                    # since evenness is the same, sort by fragmentation (low is good)
                    best_alloc_candidate_arr.append(
                        sorted(alloc_candidate_arr, key=lambda x: x[2])[-1],
                    )
                    # there is no need to look at more devices if the desired evenness is achieved
                    if max_evenness.quantize(self.digits) == self.digits:
                        best_alloc_candidate_arr = best_alloc_candidate_arr[-1:]
                        break
                # choose the best allocation with the three criteria
                slot_allocation = sorted(
                    best_alloc_candidate_arr,
                    key=operator.itemgetter(1, 2, 3),
                )[-1][0]
            allocation[slot_name] = slot_allocation
            for dev_id, value in slot_allocation.items():
                self.allocations[slot_name][dev_id] += value
            self.update_affinity_hint(slot_allocation, affinity_hint)
        return allocation

    def apply_allocation(
        self,
        existing_alloc: Mapping[SlotName, Mapping[DeviceId, Decimal]],
    ) -> None:
        for slot_name, per_device_alloc in existing_alloc.items():
            for device_id, alloc in per_device_alloc.items():
                self.allocations[slot_name][device_id] += alloc

    def free(
        self,
        existing_alloc: Mapping[SlotName, Mapping[DeviceId, Decimal]],
    ) -> None:
        for slot_name, per_device_alloc in existing_alloc.items():
            for device_id, alloc in per_device_alloc.items():
                self.allocations[slot_name][device_id] -= alloc



from collections.abc import Mapping
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Optional, override

from ai.backend.agent.exception import NetworkPluginNotFound
from ai.backend.agent.plugin.network import NetworkPluginContext
from ai.backend.common.stage.types import (
    ArgsSpecGenerator,
    Provisioner,
    ProvisionStage,
)
from ai.backend.common.types import (
    ClusterInfo,
    ClusterSSHKeyPair,
    ClusterSSHPortMapping,
    KernelCreationConfig,
)


@dataclass
class NetworkConfig:
    mode: Optional[str]
    network_name: str

    def to_dict(self) -> dict[str, Any]:
        return {
            "mode": self.mode,
            "network_name": self.network_name,
        }


@dataclass
class NetworkSpec:
    kernel_config: KernelCreationConfig

    cluster_size: int
    replicas: Mapping[str, int]  # per-role kernel counts
    network_config: NetworkConfig
    ssh_keypair: ClusterSSHKeyPair
    cluster_ssh_port_mapping: Optional[ClusterSSHPortMapping]

    gwbridge_subnet: Optional[str]
    alternative_bridge: Optional[str]


class NetworkSpecGenerator(ArgsSpecGenerator[NetworkSpec]):
    pass


@dataclass
class NetworkResult:
    container_configs: list[dict[str, Any]]


class NetworkProvisioner(Provisioner[NetworkSpec, NetworkResult]):
    def __init__(self, network_plugin_ctx: NetworkPluginContext) -> None:
        self.network_plugin_ctx = network_plugin_ctx

    @property
    @override
    def name(self) -> str:
        return "docker-network"

    @override
    async def setup(self, spec: NetworkSpec) -> NetworkResult:
        configs = await self._prepare_network_config(spec)
        return NetworkResult(
            container_configs=configs,
        )

    async def _prepare_network_config(self, spec: NetworkSpec) -> list[dict[str, Any]]:
        # FIXME: find out way to inect network ID to kernel resource spec
        base_networks = await self._prepare_base_network(spec)
        alternative_bridge_networks = await self._prepare_alternative_bridge_network(spec)
        rdma_networks = await self._prepare_rdma_network(spec)
        return [
            *base_networks,
            *alternative_bridge_networks,
            *rdma_networks,
        ]

    async def _prepare_base_network(self, spec: NetworkSpec) -> list[dict[str, Any]]:
        match spec.network_config.mode:
            case "bridge":
                return await self._prepare_bridge_network(spec)
            case mode if mode:
                return await self._prepare_plugin_network(spec, mode)
            case _:
                # TODO: handle case when no network mode is specified
                return []

    async def _prepare_bridge_network(self, spec: NetworkSpec) -> list[dict[str, Any]]:
        result: list[dict[str, Any]] = []
        result.append({
            "HostConfig": {
                "NetworkMode": spec.network_config.network_name,
            },
            "NetworkingConfig": {
                "EndpointsConfig": {
                    spec.network_config.network_name: {
                        "Aliases": [spec.kernel_config["cluster_hostname"]],
                    },
                },
            },
        })
        return result

    async def _prepare_plugin_network(self, spec: NetworkSpec, mode: str) -> list[dict[str, Any]]:
        try:
            plugin = self.network_plugin_ctx.plugins[mode]
        except KeyError:
            raise NetworkPluginNotFound(f"Network plugin {mode} not loaded!")

        cluster_info = ClusterInfo(
            mode=spec.kernel_config["cluster_mode"],
            size=spec.cluster_size,
            replicas=spec.replicas,
            network_config=spec.network_config.to_dict(),
            ssh_keypair=spec.ssh_keypair,
            cluster_ssh_port_mapping=spec.cluster_ssh_port_mapping,
        )
        container_config = await plugin.join_network(
            spec.kernel_config, cluster_info, **spec.network_config.to_dict()
        )
        result: list[dict[str, Any]] = []
        result.append(container_config)
        if spec.gwbridge_subnet is not None:
            result.append({
                "Env": [f"OMPI_MCA_btl_tcp_if_exclude=127.0.0.1/32,{spec.gwbridge_subnet}"],
            })
        return result

    async def _prepare_alternative_bridge_network(self, spec: NetworkSpec) -> list[dict[str, Any]]:
        result: list[dict[str, Any]] = []
        if spec.alternative_bridge is not None:
            result.append({
                "HostConfig": {
                    "NetworkMode": spec.alternative_bridge,
                },
            })
        return result

    async def _prepare_rdma_network(self, spec: NetworkSpec) -> list[dict[str, Any]]:
        result: list[dict[str, Any]] = []
        # RDMA mounts
        ib_root = Path("/dev/infiniband")
        if ib_root.is_dir() and (ib_root / "uverbs0").exists():
            result.append({
                "HostConfig": {
                    "Devices": [
                        {
                            "PathOnHost": "/dev/infiniband",
                            "PathInContainer": "/dev/infiniband",
                            "CgroupPermissions": "rwm",
                        },
                    ],
                },
            })
        return result

    @override
    async def teardown(self, resource: NetworkResult) -> None:
        pass


class NetworkStage(ProvisionStage[NetworkSpec, NetworkResult]):
    pass



import asyncio
import logging
import os
import signal
import ssl
import subprocess
import sys
from collections.abc import Sequence
from http import HTTPStatus
from pathlib import Path
from pprint import pformat, pprint
from typing import Any, AsyncGenerator

import aiofiles
import aiotools
import click
import trafaret as t
from aiohttp import web
from setproctitle import setproctitle

from ai.backend.common import config, utils
from ai.backend.common import validators as tx
from ai.backend.common.etcd import AsyncEtcd, ConfigScopes
from ai.backend.common.msgpack import DEFAULT_PACK_OPTS, DEFAULT_UNPACK_OPTS
from ai.backend.common.utils import Fstab
from ai.backend.logging import BraceStyleAdapter, Logger, LogLevel
from ai.backend.logging.config import LoggingConfig

__version__ = (Path(__file__).parent.parent / "VERSION").read_text().strip()

log = BraceStyleAdapter(logging.getLogger(__spec__.name))

shutdown_enabled = False


@web.middleware
async def auth_middleware(request, handler):
    token = request.headers.get("X-BackendAI-Watcher-Token", None)
    if token == request.app["token"]:
        try:
            return await handler(request)
        except FileNotFoundError as e:
            log.info(repr(e))
            message = "Agent is not loaded with systemctl."
            return web.json_response({"message": message}, status=HTTPStatus.OK)
        except Exception as e:
            log.exception(repr(e))
            raise
    log.info("invalid requested token")
    return web.HTTPForbidden()


async def handle_status(request: web.Request) -> web.Response:
    svc = request.app["config"]["watcher"]["target-service"]
    proc = await asyncio.create_subprocess_exec(
        *["sudo", "systemctl", "is-active", svc], stdout=subprocess.PIPE
    )
    if proc.stdout is not None:
        status = (await proc.stdout.read()).strip().decode()
    else:
        status = "unknown"
    await proc.wait()
    return web.json_response({
        "agent-status": status,  # maybe also "inactive", "activating"
        "watcher-status": "active",
    })


async def handle_soft_reset(request: web.Request) -> web.Response:
    svc = request.app["config"]["watcher"]["target-service"]
    proc = await asyncio.create_subprocess_exec(*["sudo", "systemctl", "reload", svc])
    await proc.wait()
    return web.json_response({
        "result": "ok",
    })


async def handle_hard_reset(request: web.Request) -> web.Response:
    svc = request.app["config"]["watcher"]["target-service"]
    proc = await asyncio.create_subprocess_exec(*["sudo", "systemctl", "stop", svc])
    await proc.wait()
    proc = await asyncio.create_subprocess_exec(*["sudo", "systemctl", "restart", "docker.service"])
    await proc.wait()
    proc = await asyncio.create_subprocess_exec(*["sudo", "systemctl", "start", svc])
    await proc.wait()
    return web.json_response({
        "result": "ok",
    })


async def handle_shutdown(request: web.Request) -> web.Response:
    global shutdown_enabled
    svc = request.app["config"]["watcher"]["target-service"]
    proc = await asyncio.create_subprocess_exec(*["sudo", "systemctl", "stop", svc])
    await proc.wait()
    shutdown_enabled = True
    signal.alarm(1)
    return web.json_response({
        "result": "ok",
    })


async def handle_agent_start(request: web.Request) -> web.Response:
    svc = request.app["config"]["watcher"]["target-service"]
    proc = await asyncio.create_subprocess_exec(*["sudo", "systemctl", "start", svc])
    await proc.wait()
    return web.json_response({
        "result": "ok",
    })


async def handle_agent_stop(request: web.Request) -> web.Response:
    svc = request.app["config"]["watcher"]["target-service"]
    proc = await asyncio.create_subprocess_exec(*["sudo", "systemctl", "stop", svc])
    await proc.wait()
    return web.json_response({
        "result": "ok",
    })


async def handle_agent_restart(request: web.Request) -> web.Response:
    svc = request.app["config"]["watcher"]["target-service"]
    proc = await asyncio.create_subprocess_exec(*["sudo", "systemctl", "restart", svc])
    await proc.wait()
    return web.json_response({
        "result": "ok",
    })


async def handle_fstab_detail(request: web.Request) -> web.Response:
    log.info("HANDLE_FSTAB_DETAIL")
    params = request.query
    fstab_path = params.get("fstab_path", "/etc/fstab")
    async with aiofiles.open(fstab_path, mode="r") as fp:
        content = await fp.read()
        return web.Response(text=content)


async def handle_list_mounts(request: web.Request) -> web.Response:
    log.info("HANDLE_LIST_MOUNT")
    config = request.app["config_server"]
    mount_prefix = await config.get("volumes/_mount")
    if mount_prefix is None:
        mount_prefix = "/mnt"
    mounts = set()
    for p in Path(mount_prefix).iterdir():
        if p.is_dir() and p.is_mount():
            mounts.add(str(p))
    return web.json_response(sorted(mounts))


async def handle_mount(request: web.Request) -> web.Response:
    log.info("HANDLE_MOUNT")
    params = await request.json()
    config = request.app["config_server"]
    mount_prefix = await config.get("volumes/_mount")
    if mount_prefix is None:
        mount_prefix = "/mnt"
    mountpoint = Path(mount_prefix) / params["name"]
    mountpoint.mkdir(exist_ok=True)
    if params.get("options", None):
        cmd = [
            "sudo",
            "mount",
            "-t",
            params["fs_type"],
            "-o",
            params["options"],
            params["fs_location"],
            str(mountpoint),
        ]
    else:
        cmd = ["sudo", "mount", "-t", params["fs_type"], params["fs_location"], str(mountpoint)]
    proc = await asyncio.create_subprocess_exec(
        *cmd, stdout=asyncio.subprocess.PIPE, stderr=asyncio.subprocess.PIPE
    )
    raw_out, raw_err = await proc.communicate()
    out = raw_out.decode("utf8")
    err = raw_err.decode("utf8")
    await proc.wait()
    if err:
        log.error("Mount error: " + err)
        return web.Response(text=err, status=HTTPStatus.INTERNAL_SERVER_ERROR)
    log.info("Mounted " + params["name"] + " on " + mount_prefix)
    if params["edit_fstab"]:
        fstab_path = params["fstab_path"] if params["fstab_path"] else "/etc/fstab"
        # FIXME: Remove ignore if https://github.com/python/typeshed/pull/4650 is released
        async with aiofiles.open(fstab_path, mode="r+") as fp:  # type: ignore
            fstab = Fstab(fp)
            await fstab.add(
                params["fs_location"], str(mountpoint), params["fs_type"], params["options"]
            )
    return web.Response(text=out)


async def handle_umount(request: web.Request) -> web.Response:
    log.info("HANDLE_UMOUNT")
    params = await request.json()
    config = request.app["config_server"]
    mount_prefix = await config.get("volumes/_mount")
    if mount_prefix is None:
        mount_prefix = "/mnt"
    mountpoint = Path(mount_prefix) / params["name"]
    assert Path(mount_prefix) != mountpoint
    proc = await asyncio.create_subprocess_exec(
        *[
            "sudo",
            "umount",
            str(mountpoint),
        ],
        stdout=asyncio.subprocess.PIPE,
        stderr=asyncio.subprocess.PIPE,
    )
    raw_out, raw_err = await proc.communicate()
    out = raw_out.decode("utf8")
    err = raw_err.decode("utf8")
    await proc.wait()
    if err:
        log.error("Unmount error: " + err)
        return web.Response(text=err, status=HTTPStatus.INTERNAL_SERVER_ERROR)
    log.info("Unmounted " + params["name"] + " from " + mount_prefix)
    try:
        mountpoint.rmdir()  # delete directory if empty
    except OSError:
        pass
    if params["edit_fstab"]:
        fstab_path = params["fstab_path"] if params["fstab_path"] else "/etc/fstab"
        # FIXME: Remove ignore if https://github.com/python/typeshed/pull/4650 is released
        async with aiofiles.open(fstab_path, mode="r+") as fp:  # type: ignore
            fstab = Fstab(fp)
            await fstab.remove_by_mountpoint(str(mountpoint))
    return web.Response(text=out)


async def init_app(app):
    r = app.router.add_route
    r("GET", "/", handle_status)
    if app["config"]["watcher"]["soft-reset-available"]:
        r("POST", "/soft-reset", handle_soft_reset)
    r("POST", "/hard-reset", handle_hard_reset)
    r("POST", "/shutdown", handle_shutdown)
    r("POST", "/agent/start", handle_agent_start)
    r("POST", "/agent/stop", handle_agent_stop)
    r("POST", "/agent/restart", handle_agent_restart)
    r("GET", "/fstab", handle_fstab_detail)
    r("GET", "/mounts", handle_list_mounts)
    r("POST", "/mounts", handle_mount)
    r("DELETE", "/mounts", handle_umount)


async def shutdown_app(app):
    pass


async def prepare_hook(request, response):
    response.headers["Server"] = "BackendAI-AgentWatcher"


@aiotools.server_context
async def watcher_server(
    loop: asyncio.AbstractEventLoop,
    pidx: int,
    args: Sequence[Any],
) -> AsyncGenerator[Any, signal.Signals]:
    global shutdown_enabled

    app = web.Application()
    app["config"] = args[0]
    logging_config: LoggingConfig = args[0]["logging"]
    log_endpoint: str = args[1]
    logger = Logger(
        logging_config,
        is_master=True,
        log_endpoint=log_endpoint,
        msgpack_options={
            "pack_opts": DEFAULT_PACK_OPTS,
            "unpack_opts": DEFAULT_UNPACK_OPTS,
        },
    )
    with logger:
        etcd_credentials = None
        if app["config"]["etcd"]["user"]:
            etcd_credentials = {
                "user": app["config"]["etcd"]["user"],
                "password": app["config"]["etcd"]["password"],
            }
        scope_prefix_map = {
            ConfigScopes.GLOBAL: "",
        }
        etcd = AsyncEtcd(
            app["config"]["etcd"]["addr"],
            app["config"]["etcd"]["namespace"],
            scope_prefix_map=scope_prefix_map,
            credentials=etcd_credentials,
        )
        app["config_server"] = etcd

        token = await etcd.get("config/watcher/token")
        if token is None:
            token = "insecure"
        log.debug("watcher authentication token: {}", token)
        app["token"] = token

        app.middlewares.append(auth_middleware)
        app.on_shutdown.append(shutdown_app)
        app.on_startup.append(init_app)
        app.on_response_prepare.append(prepare_hook)
        ssl_ctx = None
        if app["config"]["watcher"]["ssl-enabled"]:
            ssl_ctx = ssl.create_default_context(ssl.Purpose.CLIENT_AUTH)
            ssl_ctx.load_cert_chain(
                str(app["config"]["watcher"]["ssl-cert"]),
                str(app["config"]["watcher"]["ssl-privkey"]),
            )
        runner = web.AppRunner(app)
        await runner.setup()
        watcher_addr = app["config"]["watcher"]["service-addr"]
        site = web.TCPSite(
            runner,
            str(watcher_addr.host),
            watcher_addr.port,
            backlog=5,
            reuse_port=True,
            ssl_context=ssl_ctx,
        )
        await site.start()
        log.info("started at {}", watcher_addr)
        try:
            stop_sig = yield
        finally:
            log.info("shutting down...")
            if stop_sig == signal.SIGALRM and shutdown_enabled:
                log.warning("shutting down the agent node!")
                subprocess.run(["shutdown", "-h", "now"])
            await runner.cleanup()


@click.command()
@click.option(
    "-f",
    "--config-path",
    "--config",
    type=click.Path(exists=True, dir_okay=False),
    default=None,
    help="The config file path. (default: ./agent.toml and /etc/backend.ai/agent.toml)",
)
@click.option(
    "--debug",
    is_flag=True,
    help="Set the logging level to DEBUG",
)
@click.option(
    "--log-level",
    type=click.Choice([*LogLevel], case_sensitive=False),
    default=LogLevel.NOTSET,
    help="Set the logging verbosity level",
)
@click.pass_context
def main(
    ctx: click.Context,
    config_path: str,
    log_level: LogLevel,
    debug: bool,
) -> None:
    watcher_config_iv = (
        t.Dict({
            t.Key("watcher"): t.Dict({
                t.Key("service-addr", default=("0.0.0.0", 6009)): tx.HostPortPair,
                t.Key("ssl-enabled", default=False): t.Bool,
                t.Key("ssl-cert", default=None): t.Null | tx.Path(type="file"),
                t.Key("ssl-key", default=None): t.Null | tx.Path(type="file"),
                t.Key("target-service", default="backendai-agent.service"): t.String,
                t.Key("soft-reset-available", default=False): t.Bool,
            }).allow_extra("*"),
            t.Key("logging"): t.Any,  # checked in ai.backend.logging
            t.Key("debug"): t.Dict({
                t.Key("enabled", default=False): t.Bool,
            }).allow_extra("*"),
        })
        .merge(config.etcd_config_iv)
        .allow_extra("*")
    )

    raw_cfg, cfg_src_path = config.read_from_file(config_path, "agent")

    config.override_with_env(raw_cfg, ("etcd", "namespace"), "BACKEND_NAMESPACE")
    config.override_with_env(raw_cfg, ("etcd", "addr"), "BACKEND_ETCD_ADDR")
    config.override_with_env(raw_cfg, ("etcd", "user"), "BACKEND_ETCD_USER")
    config.override_with_env(raw_cfg, ("etcd", "password"), "BACKEND_ETCD_PASSWORD")
    config.override_with_env(
        raw_cfg, ("watcher", "service-addr", "host"), "BACKEND_WATCHER_SERVICE_IP"
    )
    config.override_with_env(
        raw_cfg, ("watcher", "service-addr", "port"), "BACKEND_WATCHER_SERVICE_PORT"
    )
    logging_config = LoggingConfig.model_validate(raw_cfg["logging"])
    if debug:
        log_level = LogLevel.DEBUG
    if log_level != LogLevel.NOTSET:
        logging_config.level = log_level
        logging_config.pkg_ns["ai.backend"] = log_level
        config.override_key(raw_cfg, ("debug", "enabled"), log_level == LogLevel.DEBUG)

    try:
        cfg = config.check(raw_cfg, watcher_config_iv)
        if "debug" in cfg and cfg["debug"]["enabled"]:
            print("== Watcher configuration ==")
            pprint(cfg)
        cfg["_src"] = cfg_src_path
        cfg["logging"] = logging_config
    except config.ConfigurationError as e:
        print("Validation of watcher configuration has failed:", file=sys.stderr)
        print(pformat(e.invalid_data), file=sys.stderr)
        raise click.Abort()

    # Change the filename from the logging config's file section.
    log_sockpath = Path(f"/tmp/backend.ai/ipc/watcher-logger-{os.getpid()}.sock")
    log_sockpath.parent.mkdir(parents=True, exist_ok=True)
    log_endpoint = f"ipc://{log_sockpath}"
    logger = Logger(
        logging_config,
        is_master=True,
        log_endpoint=log_endpoint,
        msgpack_options={
            "pack_opts": DEFAULT_PACK_OPTS,
            "unpack_opts": DEFAULT_UNPACK_OPTS,
        },
    )
    if "file" in logging_config.drivers:
        assert logging_config.file is not None
        fn = Path(logging_config.file.filename)
        logging_config.file.filename = f"{fn.stem}-watcher{fn.suffix}"

    setproctitle(f"backend.ai: watcher {cfg['etcd']['namespace']}")
    with logger:
        log.info("Backend.AI Agent Watcher {0}", __version__)
        log.info("runtime: {0}", utils.env_info())

        log_config = logging.getLogger("ai.backend.agent.config")
        log_config.debug("debug mode enabled.")

        aiotools.start_server(
            watcher_server,
            num_workers=1,
            args=(cfg, log_endpoint),
            stop_signals={signal.SIGINT, signal.SIGTERM, signal.SIGALRM},
        )
        log.info("exit.")



import asyncio
import enum
import uuid
from dataclasses import dataclass
from pathlib import Path
from typing import Any, Mapping, Optional, Sequence, TypeAlias

import attrs
from aiohttp.typedefs import Middleware
from pydantic import BaseModel

from ai.backend.common.docker import LabelName
from ai.backend.common.events.kernel import KernelLifecycleEventReason
from ai.backend.common.types import (
    AgentId,
    ContainerId,
    ContainerStatus,
    KernelId,
    MountTypes,
    SessionId,
)


class AgentBackend(enum.StrEnum):
    # The list of importable backend names under "ai.backend.agent" pkg namespace.
    DOCKER = "docker"
    KUBERNETES = "kubernetes"
    DUMMY = "dummy"


@attrs.define(auto_attribs=True, slots=True)
class VolumeInfo:
    name: str  # volume name
    container_path: str  # in-container path as str
    mode: str  # 'rw', 'ro', 'rwm'


@attrs.define(auto_attribs=True, slots=True)
class MountInfo:
    mode: MountTypes
    src_path: Path
    dst_path: Path


@attrs.define(auto_attribs=True, slots=True)
class Port:
    host: str
    private_port: int
    host_port: int


@attrs.define(auto_attribs=True, slots=True)
class AgentEventData:
    type: str
    data: dict[str, Any]


@attrs.define(auto_attribs=True, slots=True)
class Container:
    id: ContainerId
    status: ContainerStatus
    image: str
    labels: Mapping[str, str]
    ports: Sequence[Port]
    backend_obj: Any  # used to keep the backend-specific data

    @property
    def human_readable_id(self) -> str:
        """
        Returns a human-readable version of the container ID.
        This is useful for logging and debugging purposes.
        """
        return str(self.id)[:12]

    @property
    def kernel_id(self) -> KernelId:
        raw_kernel_id = self.labels[LabelName.KERNEL_ID]
        return KernelId(uuid.UUID(raw_kernel_id))

    @property
    def session_id(self) -> SessionId:
        raw_session_id = self.labels[LabelName.SESSION_ID]
        return SessionId(uuid.UUID(raw_session_id))


class KernelLifecycleStatus(enum.StrEnum):
    """
    The lifecycle status of `AbstractKernel` object.

    By default, the state of a newly created kernel is `PREPARING`.
    The state of a kernel changes from `PREPARING` to `RUNNING` after the kernel starts a container successfully.
    It changes from `RUNNING` to `TERMINATING` before destroy kernel.
    """

    PREPARING = enum.auto()
    RUNNING = enum.auto()
    TERMINATING = enum.auto()


class LifecycleEvent(enum.IntEnum):
    DESTROY = 0
    CLEAN = 1
    START = 2


@attrs.define(auto_attribs=True, slots=True)
class ContainerLifecycleEvent:
    kernel_id: KernelId
    session_id: SessionId
    container_id: Optional[ContainerId]
    event: LifecycleEvent
    reason: KernelLifecycleEventReason
    done_future: Optional[asyncio.Future] = None
    exit_code: Optional[int] = None
    suppress_events: bool = False

    def __str__(self):
        if self.container_id:
            cid = self.container_id[:13]
        else:
            cid = "unknown"
        return (
            "LifecycleEvent("
            f"{self.event.name}, "
            f"k:{self.kernel_id}, "
            f"c:{cid}, "
            f"reason:{self.reason!r})"
        )

    def set_done_future_result(self, result: Any):
        if self.done_future is not None:
            try:
                self.done_future.set_result(result)
            except asyncio.InvalidStateError:
                # The future is already done, ignore the error
                pass

    def set_done_future_exception(self, exception: Exception):
        if self.done_future is not None:
            try:
                self.done_future.set_exception(exception)
            except asyncio.InvalidStateError:
                # The future is already done, ignore the error
                pass


@dataclass
class KernelOwnershipData:
    kernel_id: KernelId
    session_id: SessionId
    agent_id: AgentId
    owner_user_id: Optional[uuid.UUID] = None
    owner_project_id: Optional[uuid.UUID] = None

    def __post_init__(self):
        def to_uuid(value: Optional[str]) -> Optional[uuid.UUID]:
            return uuid.UUID(value) if value is not None else None

        self.owner_user_id = to_uuid(self.owner_user_id)
        self.owner_project_id = to_uuid(self.owner_project_id)

    @property
    def owner_user_id_to_str(self) -> Optional[str]:
        return str(self.owner_user_id) if self.owner_user_id is not None else None

    @property
    def owner_project_id_to_str(self) -> Optional[str]:
        return str(self.owner_project_id) if self.owner_project_id is not None else None


WebMiddleware: TypeAlias = Middleware


class HealthResponse(BaseModel):
    """Standard health check response"""

    status: str
    version: str
    component: str



import logging
from typing import TYPE_CHECKING, Optional, override

from ai.backend.common.events.dispatcher import EventProducer
from ai.backend.common.events.event_types.agent.anycast import (
    AgentStatusHeartbeat,
    ContainerStatusData,
)
from ai.backend.common.observer.types import AbstractObserver
from ai.backend.common.types import ContainerId, ContainerStatus, KernelContainerId
from ai.backend.logging.utils import BraceStyleAdapter

if TYPE_CHECKING:
    from ..agent import AbstractAgent


log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class HeartbeatObserver(AbstractObserver):
    def __init__(
        self,
        agent: "AbstractAgent",
        event_producer: EventProducer,
    ) -> None:
        self._agent = agent
        self._event_producer = event_producer

    @property
    @override
    def name(self) -> str:
        return "agent_hearbeat"

    @override
    async def observe(self) -> None:
        containers = await self._agent.enumerate_containers(ContainerStatus.active_set())

        container_data = [
            ContainerStatusData(
                container.id,
                kernel_id,
                container.status,
            )
            for kernel_id, container in containers
        ]
        kernel_data = [
            KernelContainerId(
                kernel_id,
                ContainerId(kernel_obj.container_id) if kernel_obj.container_id else None,
            )
            for kernel_id, kernel_obj in self._agent.kernel_registry.items()
        ]
        await self._event_producer.anycast_event(
            AgentStatusHeartbeat(
                self._agent.id,
                container_data,
                kernel_data,
            ),
        )

    @override
    def observe_interval(self) -> float:
        return 60.0

    @classmethod
    @override
    def timeout(cls) -> Optional[float]:
        return 10.0

    @override
    async def cleanup(self) -> None:
        pass



import asyncio
import os
from collections.abc import Iterable
from dataclasses import dataclass
from pathlib import Path
from typing import Optional, override

from ai.backend.common.docker import KernelFeatures
from ai.backend.common.json import dump_json
from ai.backend.common.stage.types import (
    ArgsSpecGenerator,
    Provisioner,
    ProvisionStage,
)
from ai.backend.common.types import (
    ClusterSSHKeyPair,
    ClusterSSHPortMapping,
)


@dataclass
class SSHSpec:
    config_dir: Path
    ssh_keypair: Optional[ClusterSSHKeyPair]
    cluster_ssh_port_mapping: Optional[ClusterSSHPortMapping]

    agent_kernel_features: frozenset[KernelFeatures]
    agent_kernel_uid: int
    agent_kernel_gid: int

    overriding_uid: Optional[int]
    overriding_gid: Optional[int]
    supplementary_gids: set[int]


class SSHSpecGenerator(ArgsSpecGenerator[SSHSpec]):
    pass


@dataclass
class SSHResult:
    pub_key_path: Optional[Path]
    priv_key_path: Optional[Path]
    port_mapping_json_path: Optional[Path]


class Chowner:
    def __init__(self, spec: SSHSpec) -> None:
        self._spec = spec
        self._euid = os.geteuid()

    def chown_paths_if_root(
        self, paths: Iterable[Path], uid: Optional[int], gid: Optional[int]
    ) -> None:
        if self._euid == 0:  # only possible when I am root.
            for p in paths:
                if KernelFeatures.UID_MATCH in self._spec.agent_kernel_features:
                    valid_uid = uid if uid is not None else self._spec.agent_kernel_uid
                    valid_gid = gid if gid is not None else self._spec.agent_kernel_gid
                else:
                    stat = os.stat(p)
                    valid_uid = uid if uid is not None else stat.st_uid
                    valid_gid = gid if gid is not None else stat.st_gid
                os.chown(p, valid_uid, valid_gid)


class SSHProvisioner(Provisioner[SSHSpec, SSHResult]):
    @property
    @override
    def name(self) -> str:
        return "docker-ssh"

    @override
    async def setup(self, spec: SSHSpec) -> SSHResult:
        return await self._write_config(spec)

    async def _write_config(self, spec: SSHSpec) -> SSHResult:
        loop = asyncio.get_running_loop()
        return await loop.run_in_executor(None, self._write_config_func, spec)

    def _write_config_func(self, spec: SSHSpec) -> SSHResult:
        sshkey = spec.ssh_keypair
        if sshkey is None:
            return SSHResult(None, None, None)

        priv_key_path = spec.config_dir / "ssh" / "id_cluster"
        pub_key_path = spec.config_dir / "ssh" / "id_cluster.pub"
        priv_key_path.parent.mkdir(parents=True, exist_ok=True)
        priv_key_path.write_text(sshkey["private_key"])
        pub_key_path.write_text(sshkey["public_key"])

        do_override = False
        if (ouid := spec.overriding_uid) is not None:
            do_override = True

        if (ogid := spec.overriding_gid) is not None:
            do_override = True

        chowner = Chowner(spec)
        if do_override:
            chowner.chown_paths_if_root([priv_key_path, pub_key_path], ouid, ogid)
        else:
            if KernelFeatures.UID_MATCH in spec.agent_kernel_features:
                chowner.chown_paths_if_root(
                    [priv_key_path, pub_key_path], spec.agent_kernel_uid, spec.agent_kernel_gid
                )

        priv_key_path.chmod(0o600)
        if cluster_ssh_port_mapping := spec.cluster_ssh_port_mapping:
            port_mapping_json_path = spec.config_dir / "ssh" / "port-mapping.json"
            port_mapping_json_path.write_bytes(dump_json(cluster_ssh_port_mapping))
        else:
            port_mapping_json_path = None
        return SSHResult(
            pub_key_path=pub_key_path,
            priv_key_path=priv_key_path,
            port_mapping_json_path=port_mapping_json_path,
        )

    @override
    async def teardown(self, resource: SSHResult) -> None:
        pass


class SSHStage(ProvisionStage[SSHSpec, SSHResult]):
    pass



import logging
from decimal import Decimal
from pathlib import Path
from typing import Any, Mapping, MutableMapping, Optional

import aiofiles

from ai.backend.common.etcd import AsyncEtcd
from ai.backend.common.types import DeviceName, SlotName
from ai.backend.logging import BraceStyleAdapter

from ..exception import InitializationError
from ..resources import (
    AbstractComputePlugin,
    ComputePluginContext,
    KernelResourceSpec,
    known_slot_types,
)

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


async def load_resources(
    etcd: AsyncEtcd, local_config: Mapping[str, Any]
) -> Mapping[DeviceName, AbstractComputePlugin]:
    """
    Detect and load the accelerator plugins.

    limit_cpus, limit_gpus are deprecated.
    """

    compute_device_types: MutableMapping[DeviceName, AbstractComputePlugin] = {}

    # Initialize intrinsic plugins by ourselves.
    from .intrinsic import CPUPlugin, MemoryPlugin

    compute_plugin_ctx = ComputePluginContext(
        etcd,
        local_config,
    )
    await compute_plugin_ctx.init(
        allowlist=local_config["agent"]["allow-compute-plugins"],
        blocklist=local_config["agent"]["block-compute-plugins"],
    )
    if "cpu" not in compute_plugin_ctx.plugins:
        cpu_config = await etcd.get_prefix("config/plugins/cpu")
        cpu_plugin = CPUPlugin(cpu_config, local_config)
        compute_plugin_ctx.attach_intrinsic_device(cpu_plugin)
    if "mem" not in compute_plugin_ctx.plugins:
        memory_config = await etcd.get_prefix("config/plugins/memory")
        memory_plugin = MemoryPlugin(memory_config, local_config)
        compute_plugin_ctx.attach_intrinsic_device(memory_plugin)
    for plugin_name, plugin_instance in compute_plugin_ctx.plugins.items():
        if not all(
            (invalid_name := sname, sname.startswith(f"{plugin_instance.key}."))[1]
            for sname, _ in plugin_instance.slot_types
            if sname not in {"cpu", "mem"}
        ):
            raise InitializationError(
                "Slot types defined by an accelerator plugin must be prefixed by the plugin's key.",
                invalid_name,  # noqa: F821
                plugin_instance.key,
            )
        if plugin_instance.key in compute_device_types:
            raise InitializationError(
                f"A plugin defining the same key '{plugin_instance.key}' already exists. "
                "You may need to uninstall it first."
            )
        compute_device_types[plugin_instance.key] = plugin_instance

    return compute_device_types


async def scan_available_resources(
    local_config: Mapping[str, Any],
    compute_device_types: Mapping[DeviceName, AbstractComputePlugin],
) -> Mapping[SlotName, Decimal]:
    """
    Detect available computing resource of the system.
    """
    slots: MutableMapping[SlotName, Decimal] = {}
    for key, computer in compute_device_types.items():
        known_slot_types.update(computer.slot_types)  # type: ignore  # (only updated here!)
        resource_slots = await computer.available_slots()
        for sname, sval in resource_slots.items():
            slots[sname] = Decimal(sval)
            if slots[sname] <= 0 and sname in (SlotName("cpu"), SlotName("mem")):
                raise InitializationError(
                    f"The resource slot '{sname}' is not sufficient (zero or below zero). "
                    "Try to adjust the reserved resources or use a larger machine."
                )

    log.info("Resource slots: {!r}", slots)
    log.info("Slot types: {!r}", known_slot_types)
    return slots


async def get_resource_spec_from_container(container_info) -> Optional[KernelResourceSpec]:
    for mount in container_info["HostConfig"]["Mounts"]:
        if mount["Target"] == "/home/config":
            async with aiofiles.open(Path(mount["Source"]) / "resource.txt", "r") as f:  # type: ignore
                return await KernelResourceSpec.aread_from_file(f)
    else:
        return None



"""
Bootstrap stage for kernel lifecycle.

This stage handles creation of bootstrap scripts in the container work directory.
"""

import asyncio
from dataclasses import dataclass
from pathlib import Path
from typing import Optional, override

from ai.backend.common.stage.types import ArgsSpecGenerator, Provisioner, ProvisionStage

from .utils import ChownUtil, PathOwnerDeterminer


@dataclass
class AgentConfig:
    kernel_features: frozenset[str]
    kernel_uid: int
    kernel_gid: int


@dataclass
class BootstrapSpec:
    work_dir: Path
    bootstrap_script: Optional[str]

    # Override UID/GID settings
    uid_override: Optional[int]
    gid_override: Optional[int]

    agent_config: AgentConfig


class BootstrapSpecGenerator(ArgsSpecGenerator[BootstrapSpec]):
    pass


@dataclass
class BootstrapResult:
    bootstrap_path: Optional[Path]


class BootstrapProvisioner(Provisioner[BootstrapSpec, BootstrapResult]):
    """
    Provisioner for bootstrap script creation.

    Creates bootstrap.sh script in the work directory if provided in kernel config.
    """

    @property
    @override
    def name(self) -> str:
        return "docker-bootstrap"

    @override
    async def setup(self, spec: BootstrapSpec) -> BootstrapResult:
        loop = asyncio.get_running_loop()

        bootstrap_path = await loop.run_in_executor(None, self._write_bootstrap_script, spec)
        return BootstrapResult(bootstrap_path=bootstrap_path)

    def _write_bootstrap_script(self, spec: BootstrapSpec) -> Optional[Path]:
        if not spec.bootstrap_script:
            return None

        bootstrap_path = spec.work_dir / "bootstrap.sh"
        bootstrap_path.write_text(spec.bootstrap_script)

        # Set proper ownership
        owner_determiner = PathOwnerDeterminer.by_kernel_features(
            spec.agent_config.kernel_uid,
            spec.agent_config.kernel_gid,
            spec.agent_config.kernel_features,
        )
        final_uid, final_gid = owner_determiner.determine(
            bootstrap_path, uid_override=spec.uid_override, gid_override=spec.gid_override
        )
        ChownUtil().chown_path(bootstrap_path, final_uid, final_gid)
        return bootstrap_path

    @override
    async def teardown(self, resource: BootstrapResult) -> None:
        # Bootstrap script is cleaned up with scratch directory
        pass


class BootstrapStage(ProvisionStage[BootstrapSpec, BootstrapResult]):
    """
    Stage for creating bootstrap scripts in kernel containers.
    """

    pass



from decimal import Decimal
from pathlib import Path
from typing import Any, Collection, Mapping, Optional, Sequence

import aiodocker

from ai.backend.agent.types import MountInfo
from ai.backend.common.types import (
    AcceleratorMetadata,
    DeviceId,
    DeviceModelInfo,
    DeviceName,
    SlotName,
    SlotTypes,
)

from .. import __version__
from ..alloc_map import AllocationStrategy
from ..resources import (
    AbstractAllocMap,
    AbstractComputeDevice,
    AbstractComputePlugin,
    DeviceSlotInfo,
    DiscretePropertyAllocMap,
)
from ..stats import (
    ContainerMeasurement,
    NodeMeasurement,
    ProcessMeasurement,
    StatContext,
)
from .agent import Container


class CPUDevice(AbstractComputeDevice):
    pass


class CPUPlugin(AbstractComputePlugin):
    """
    Represents the CPU.
    """

    resource_config: Mapping[str, Any]

    config_watch_enabled = False

    key = DeviceName("cpu")
    slot_types = [
        (SlotName("cpu"), SlotTypes.COUNT),
    ]

    def __init__(
        self,
        plugin_config: Mapping[str, Any],
        local_config: Mapping[str, Any],
        dummy_config: Mapping[str, Any],
    ) -> None:
        super().__init__(plugin_config, local_config)
        self.resource_config = dummy_config["agent"]["resource"]

    async def init(self, context: Optional[Any] = None) -> None:
        pass

    async def cleanup(self) -> None:
        pass

    async def update_plugin_config(self, new_plugin_config: Mapping[str, Any]) -> None:
        pass

    def get_metadata(self) -> AcceleratorMetadata:
        return {
            "slot_name": "cpu",
            "description": "CPU",
            "human_readable_name": "CPU",
            "display_unit": "Core",
            "number_format": {"binary": False, "round_length": 0},
            "display_icon": "cpu",
        }

    async def list_devices(self) -> Collection[AbstractComputeDevice]:
        num_core: int = self.resource_config["cpu"]["num-core"]
        return [
            CPUDevice(
                device_id=DeviceId(str(core_idx)),
                hw_location="root",
                numa_node=None,
                memory_size=0,
                processing_units=1,
            )
            for core_idx in range(num_core)
        ]

    async def available_slots(self) -> Mapping[SlotName, Decimal]:
        devices = await self.list_devices()
        return {
            SlotName("cpu"): Decimal(sum(dev.processing_units for dev in devices)),
        }

    def get_version(self) -> str:
        return __version__

    async def extra_info(self) -> Mapping[str, str]:
        return {}

    async def gather_node_measures(self, ctx: StatContext) -> Sequence[NodeMeasurement]:
        return []

    async def gather_container_measures(
        self,
        ctx: StatContext,
        container_ids: Sequence[str],
    ) -> Sequence[ContainerMeasurement]:
        return []

    async def gather_process_measures(
        self, ctx: StatContext, pid_map: Mapping[int, str]
    ) -> Sequence[ProcessMeasurement]:
        return []

    async def create_alloc_map(self) -> "AbstractAllocMap":
        devices = await self.list_devices()
        return DiscretePropertyAllocMap(
            device_slots={
                dev.device_id: DeviceSlotInfo(
                    SlotTypes.COUNT, SlotName("cpu"), Decimal(dev.processing_units)
                )
                for dev in devices
            },
        )

    async def get_hooks(self, distro: str, arch: str) -> Sequence[Path]:
        return []

    async def generate_docker_args(
        self,
        docker: aiodocker.docker.Docker,
        device_alloc,
    ) -> Mapping[str, Any]:
        cores = [*map(int, device_alloc["cpu"].keys())]
        sorted_core_ids = [*map(str, sorted(cores))]
        return {
            "HostConfig": {
                "CpuPeriod": 100_000,  # docker default
                "CpuQuota": int(100_000 * len(cores)),
                "Cpus": ",".join(sorted_core_ids),
                "CpusetCpus": ",".join(sorted_core_ids),
                # 'CpusetMems': f'{resource_spec.numa_node}',
            },
        }

    async def restore_from_container(
        self,
        container: Container,
        alloc_map: AbstractAllocMap,
    ) -> None:
        return None

    async def get_attached_devices(
        self,
        device_alloc: Mapping[SlotName, Mapping[DeviceId, Decimal]],
    ) -> Sequence[DeviceModelInfo]:
        device_ids = [*device_alloc[SlotName("cpu")].keys()]
        available_devices = await self.list_devices()
        attached_devices: list[DeviceModelInfo] = []
        for device in available_devices:
            if device.device_id in device_ids:
                attached_devices.append({
                    "device_id": device.device_id,
                    "model_name": "",
                    "data": {"cores": len(device_ids)},
                })
        return attached_devices

    async def get_docker_networks(
        self, device_alloc: Mapping[SlotName, Mapping[DeviceId, Decimal]]
    ) -> list[str]:
        return []

    async def generate_mounts(
        self, source_path: Path, device_alloc: Mapping[SlotName, Mapping[DeviceId, Decimal]]
    ) -> list[MountInfo]:
        return []


class MemoryDevice(AbstractComputeDevice):
    pass


class MemoryPlugin(AbstractComputePlugin):
    """
    Represents the main memory.
    """

    resource_config: Mapping[str, Any]

    config_watch_enabled = False

    key = DeviceName("mem")
    slot_types = [
        (SlotName("mem"), SlotTypes.BYTES),
    ]

    def __init__(
        self,
        plugin_config: Mapping[str, Any],
        local_config: Mapping[str, Any],
        dummy_config: Mapping[str, Any],
    ) -> None:
        super().__init__(plugin_config, local_config)
        self.resource_config = dummy_config["agent"]["resource"]

    async def init(self, context: Optional[Any] = None) -> None:
        pass

    async def cleanup(self) -> None:
        pass

    async def update_plugin_config(self, new_plugin_config: Mapping[str, Any]) -> None:
        pass

    def get_metadata(self) -> AcceleratorMetadata:
        return {
            "slot_name": "ram",
            "description": "Memory",
            "human_readable_name": "RAM",
            "display_unit": "GiB",
            "number_format": {"binary": True, "round_length": 0},
            "display_icon": "cpu",
        }

    async def list_devices(self) -> Collection[AbstractComputeDevice]:
        memory_size = self.resource_config["memory"]["size"]
        return [
            MemoryDevice(
                device_id=DeviceId("root"),
                device_name=self.key,
                hw_location="root",
                numa_node=0,  # the kernel setting will do the job.
                memory_size=memory_size,
                processing_units=0,
            ),
        ]

    async def available_slots(self) -> Mapping[SlotName, Decimal]:
        devices = await self.list_devices()
        return {
            SlotName("mem"): Decimal(sum(dev.memory_size for dev in devices)),
        }

    def get_version(self) -> str:
        return __version__

    async def extra_info(self) -> Mapping[str, str]:
        return {}

    async def gather_node_measures(self, ctx: StatContext) -> Sequence[NodeMeasurement]:
        return []

    async def gather_container_measures(
        self,
        ctx: StatContext,
        container_ids: Sequence[str],
    ) -> Sequence[ContainerMeasurement]:
        return []

    async def gather_process_measures(
        self, ctx: StatContext, pid_map: Mapping[int, str]
    ) -> Sequence[ProcessMeasurement]:
        return []

    async def create_alloc_map(self) -> "AbstractAllocMap":
        devices = await self.list_devices()
        return DiscretePropertyAllocMap(
            allocation_strategy=AllocationStrategy.FILL,
            device_slots={
                dev.device_id: DeviceSlotInfo(
                    SlotTypes.BYTES, SlotName("mem"), Decimal(dev.memory_size)
                )
                for dev in devices
            },
        )

    async def get_hooks(self, distro: str, arch: str) -> Sequence[Path]:
        return []

    async def generate_docker_args(
        self,
        docker: aiodocker.docker.Docker,
        device_alloc,
    ) -> Mapping[str, Any]:
        return {}

    async def restore_from_container(
        self,
        container: Container,
        alloc_map: AbstractAllocMap,
    ) -> None:
        return None

    async def get_attached_devices(
        self,
        device_alloc: Mapping[SlotName, Mapping[DeviceId, Decimal]],
    ) -> Sequence[DeviceModelInfo]:
        device_ids = [*device_alloc[SlotName("mem")].keys()]
        available_devices = await self.list_devices()
        attached_devices: list[DeviceModelInfo] = []
        for device in available_devices:
            if device.device_id in device_ids:
                attached_devices.append({
                    "device_id": device.device_id,
                    "model_name": "",
                    "data": {},
                })
        return attached_devices

    async def get_docker_networks(
        self, device_alloc: Mapping[SlotName, Mapping[DeviceId, Decimal]]
    ) -> list[str]:
        return []

    async def generate_mounts(
        self, source_path: Path, device_alloc: Mapping[SlotName, Mapping[DeviceId, Decimal]]
    ) -> list[MountInfo]:
        return []



import asyncio
import secrets
import sys
from dataclasses import dataclass
from functools import partial
from pathlib import Path
from typing import override

from aiodocker.docker import Docker

from ai.backend.agent.proxy import DomainSocketProxy, proxy_connection
from ai.backend.agent.resources import Mount
from ai.backend.agent.types import VolumeInfo
from ai.backend.agent.utils import closing_async
from ai.backend.common.docker import ImageRef
from ai.backend.common.stage.types import (
    ArgsSpecGenerator,
    Provisioner,
    ProvisionStage,
)
from ai.backend.common.types import (
    MountPermission,
    MountTypes,
)


@dataclass
class CoreDumpConfig:
    enabled: bool
    path: Path
    core_path: Path


@dataclass
class IntrinsicMountSpec:
    config_dir: Path
    work_dir: Path
    tmp_dir: Path
    scratch_type: str

    agent_sockpath: Path
    image_ref: ImageRef
    coredump: CoreDumpConfig
    ipc_base_path: Path
    domain_socket_proxies: list[DomainSocketProxy]


class IntrinsicMountSpecGenerator(ArgsSpecGenerator[IntrinsicMountSpec]):
    pass


@dataclass
class IntrinsicMountResult:
    mounts: list[Mount]
    domain_socket_proxies: list[DomainSocketProxy]


class IntrinsicMountProvisioner(Provisioner[IntrinsicMountSpec, IntrinsicMountResult]):
    """
    Provisioner for the kernel creation setup stage.
    This is a no-op provisioner as it does not create any resources.
    """

    @property
    @override
    def name(self) -> str:
        return "docker-intrinsic-mount"

    @override
    async def setup(self, spec: IntrinsicMountSpec) -> IntrinsicMountResult:
        scratch_mounts = await self._prepare_scratch_mounts(spec)
        timezone_mounts = await self._prepare_timezone_mounts(spec)
        lxcfs_mounts = await self._prepare_lxcfs_mounts(spec)
        coredump_mounts = await self._prepare_coredump_mounts(spec)
        agent_socket_mounts = await self._prepare_agent_socket_mounts(spec)
        domain_socket_proxies, domain_socket_mounts = await self._prepare_domain_socket_proxies(
            spec
        )
        mounts = [
            *scratch_mounts,
            *timezone_mounts,
            *lxcfs_mounts,
            *coredump_mounts,
            *agent_socket_mounts,
            *domain_socket_mounts,
        ]
        return IntrinsicMountResult(
            mounts=mounts,
            domain_socket_proxies=domain_socket_proxies,
        )

    async def _prepare_scratch_mounts(self, spec: IntrinsicMountSpec) -> list[Mount]:
        # scratch/config/tmp mounts
        mounts: list[Mount] = [
            Mount(
                MountTypes.BIND, spec.config_dir, Path("/home/config"), MountPermission.READ_ONLY
            ),
            Mount(MountTypes.BIND, spec.work_dir, Path("/home/work"), MountPermission.READ_WRITE),
        ]
        if sys.platform.startswith("linux") and spec.scratch_type == "memory":
            mounts.append(
                Mount(
                    MountTypes.BIND,
                    spec.tmp_dir,
                    Path("/tmp"),
                    MountPermission.READ_WRITE,
                )
            )
        return mounts

    async def _prepare_timezone_mounts(self, spec: IntrinsicMountSpec) -> list[Mount]:
        # /etc/localtime and /etc/timezone mounts
        mounts: list[Mount] = []
        if sys.platform.startswith("linux"):
            localtime_file = Path("/etc/localtime")
            timezone_file = Path("/etc/timezone")
            if localtime_file.exists():
                mounts.append(
                    Mount(
                        type=MountTypes.BIND,
                        source=localtime_file,
                        target=localtime_file,
                        permission=MountPermission.READ_ONLY,
                    )
                )
            if timezone_file.exists():
                mounts.append(
                    Mount(
                        type=MountTypes.BIND,
                        source=timezone_file,
                        target=timezone_file,
                        permission=MountPermission.READ_ONLY,
                    )
                )
        return mounts

    async def _prepare_lxcfs_mounts(self, spec: IntrinsicMountSpec) -> list[Mount]:
        # lxcfs mounts
        mounts: list[Mount] = []
        lxcfs_root = Path("/var/lib/lxcfs")
        if lxcfs_root.is_dir():
            mounts.extend(
                Mount(
                    MountTypes.BIND,
                    lxcfs_proc_path,
                    "/" / lxcfs_proc_path.relative_to(lxcfs_root),
                    MountPermission.READ_WRITE,
                )
                for lxcfs_proc_path in (lxcfs_root / "proc").iterdir()
                if lxcfs_proc_path.stat().st_size > 0
            )
            mounts.extend(
                Mount(
                    MountTypes.BIND,
                    lxcfs_root / path,
                    "/" / Path(path),
                    MountPermission.READ_WRITE,
                )
                for path in [
                    "sys/devices/system/cpu",
                    "sys/devices/system/cpu/online",
                ]
                if Path(lxcfs_root / path).exists()
            )
        return mounts

    async def _prepare_coredump_mounts(self, spec: IntrinsicMountSpec) -> list[Mount]:
        """Prepare mounts for coredump if enabled."""
        # debug mounts
        mounts: list[Mount] = []
        if spec.coredump.enabled:
            mounts.append(
                Mount(
                    MountTypes.BIND,
                    spec.coredump.path,
                    spec.coredump.core_path,
                    MountPermission.READ_WRITE,
                )
            )
        return mounts

    async def _prepare_agent_socket_mounts(self, spec: IntrinsicMountSpec) -> list[Mount]:
        # agent-socket mount
        mounts: list[Mount] = []
        if sys.platform != "darwin":
            mounts.append(
                Mount(
                    MountTypes.BIND,
                    spec.agent_sockpath,
                    Path("/opt/kernel/agent.sock"),
                    MountPermission.READ_WRITE,
                )
            )
        return mounts

    async def _prepare_domain_socket_proxies(
        self, spec: IntrinsicMountSpec
    ) -> tuple[list[DomainSocketProxy], list[Mount]]:
        loop = asyncio.get_running_loop()
        ipc_base_path = spec.ipc_base_path
        domain_socket_proxies = []
        mounts = []

        # domain-socket proxy mount
        # (used for special service containers such image importer)
        if spec.domain_socket_proxies:
            await loop.run_in_executor(
                None, partial((ipc_base_path / "proxy").mkdir, parents=True, exist_ok=True)
            )
        for proxy in spec.domain_socket_proxies:
            host_proxy_path = ipc_base_path / "proxy" / f"{secrets.token_hex(12)}.sock"
            proxy_server = await asyncio.start_unix_server(
                partial(proxy_connection, proxy.host_sock_path), str(host_proxy_path)
            )
            await loop.run_in_executor(None, host_proxy_path.chmod, 0o666)
            domain_socket_proxies.append(
                DomainSocketProxy(
                    Path(proxy.host_sock_path),
                    host_proxy_path,
                    proxy_server,
                )
            )
            mounts.append(
                Mount(
                    MountTypes.BIND,
                    host_proxy_path,
                    proxy.host_sock_path,
                    MountPermission.READ_WRITE,
                )
            )
        return domain_socket_proxies, mounts

    async def _prepare_deeplearning_sample_mounts(self, spec: IntrinsicMountSpec) -> list[Mount]:
        """
        NOTE: !! This is deprecated and should not be used in new code !!
        Prepare mounts for deeplearning sample data if needed.
        """
        deeplearning_image_keys = {
            "tensorflow",
            "caffe",
            "keras",
            "torch",
            "mxnet",
            "theano",
        }

        deeplearning_sample_volume = VolumeInfo(
            "deeplearning-samples",
            "/home/work/samples",
            "ro",
        )
        short_image_ref = spec.image_ref.short
        async with closing_async(Docker()) as docker:
            avail_volumes = (await docker.volumes.list())["Volumes"]
            if not avail_volumes:
                return []
            avail_volume_names = set(v["Name"] for v in avail_volumes)

            # deeplearning specialization
            # TODO: extract as config
            volume_list: list[VolumeInfo] = []
            for k in deeplearning_image_keys:
                if k in short_image_ref:
                    volume_list.append(deeplearning_sample_volume)
                    break

            # Mount only actually existing volumes
            volume_mount_list: list[VolumeInfo] = []
            for vol in volume_list:
                if vol.name in avail_volume_names:
                    volume_mount_list.append(vol)
            return [
                Mount(
                    MountTypes.VOLUME, Path(v.name), Path(v.container_path), MountPermission(v.mode)
                )
                for v in volume_mount_list
            ]

    @override
    async def teardown(self, resource: IntrinsicMountResult) -> None:
        pass


class IntrinsicMountStage(ProvisionStage[IntrinsicMountSpec, IntrinsicMountResult]):
    pass



import logging
from abc import ABC, abstractmethod
from dataclasses import dataclass

from ai.backend.logging import BraceStyleAdapter

from .kernel_registry import KernelRegistry

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


SAVE_COOL_DOWN_SECONDS = 60


@dataclass
class KernelRegistrySaveMetadata:
    force: bool


class AbstractKernelRegistryRecovery(ABC):
    """
    Recovery interface for loading and saving KernelRegistry
    """

    @abstractmethod
    async def load_kernel_registry(self) -> KernelRegistry:
        """
        Load the KernelRegistry from persistent storage.
        Raises: KernelRegistryNotFound
        Returns: The loaded KernelRegistry.
        """
        pass

    @abstractmethod
    async def save_kernel_registry(
        self, registry: KernelRegistry, metadata: KernelRegistrySaveMetadata
    ) -> None:
        """
        Save the KernelRegistry to persistent storage.
        args:
            registry: The KernelRegistry to save.
            metadata: Additional metadata for saving.
        Returns: None
        """
        pass



from dataclasses import dataclass
from decimal import Decimal
from typing import Optional, Sequence

from aiohttp import web

from ai.backend.common.exception import (
    BackendAIError,
    ErrorCode,
    ErrorDetail,
    ErrorDomain,
    ErrorOperation,
)
from ai.backend.common.types import DeviceId, SlotName


class InitializationError(Exception):
    """
    Errors during agent initialization and compute plugin setup
    """

    pass


class ResourceError(ValueError):
    pass


class InvalidArgumentError(RuntimeError):
    pass


class UnsupportedResource(ResourceError):
    pass


class InvalidResourceCombination(ResourceError):
    pass


class InvalidResourceArgument(ResourceError):
    pass


class NotMultipleOfQuantum(InvalidResourceArgument):
    pass


@dataclass
class InsufficientResource(ResourceError):
    msg: str
    slot_name: SlotName
    requested_alloc: Decimal
    total_allocatable: Decimal | int
    allocation: dict[SlotName, dict[DeviceId, Decimal]]
    context_tag: Optional[str] = None

    def __str__(self) -> str:
        return (
            f"InsufficientResource: {self.msg} ({self.slot_name}"
            + (f" (tag: {self.context_tag!r}), " if self.context_tag else ", ")
            + f"allocating {self.requested_alloc} out of {self.total_allocatable})"
        )

    def __reduce__(self):
        return (
            self.__class__,
            (
                self.msg,
                self.slot_name,
                self.requested_alloc,
                self.total_allocatable,
                self.allocation,
                self.context_tag,
            ),
        )


@dataclass
class FractionalResourceFragmented(ResourceError):
    msg: str
    slot_name: SlotName
    requested_alloc: Decimal
    dev_allocs: Sequence[tuple[DeviceId, Decimal]]
    context_tag: Optional[str] = None

    def __str__(self) -> str:
        return (
            f"FractionalResourceFragmented: {self.msg} ({self.slot_name}"
            + (f" (tag: {self.context_tag!r}), " if self.context_tag else ", ")
            + f"allocating {self.requested_alloc} from {self.dev_allocs})"
        )

    def __reduce__(self):
        return (
            self.__class__,
            (
                self.msg,
                self.slot_name,
                self.requested_alloc,
                self.dev_allocs,
                self.context_tag,
            ),
        )


class UnsupportedBaseDistroError(RuntimeError):
    pass


class ContainerCreationError(Exception):
    container_id: str

    def __init__(self, container_id: str, message: str | None = None, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.container_id = container_id
        self.message = message


class K8sError(Exception):
    def __init__(self, message):
        super().__init__(message)
        self.message = message


class AgentError(RuntimeError):
    """
    A dummy exception class to distinguish agent-side errors passed via
    aiozmq.rpc calls.

    It carrise two args tuple: the exception type and exception arguments from
    the agent.
    """

    def __init__(self, *args, exc_repr: Optional[str] = None):
        super().__init__(*args)
        self.exc_repr = exc_repr


class InvalidSocket(Exception):
    pass


class NetworkPluginNotFound(RuntimeError):
    pass


class ServicePortAlreadyUsedError(RuntimeError):
    """
    Raised when a port is already used by another service.
    """


class InvalidModelConfigurationError(RuntimeError):
    """
    Raised when a model configuration is invalid.
    """


class KernelRegistryNotFound(BackendAIError, web.HTTPNotFound):
    error_type = "https://backend.ai/errors/agent/kernel-registry-not-found"
    error_title = "Kernel Registry Not Found"

    @classmethod
    def error_code(cls) -> ErrorCode:
        return ErrorCode(
            domain=ErrorDomain.KERNEL_REGISTRY,
            operation=ErrorOperation.READ,
            error_detail=ErrorDetail.NOT_FOUND,
        )



from pathlib import Path

import trafaret as t

from ai.backend.common import validators as tx

DEFAULT_CONFIG_PATH = Path.cwd() / "agent.dummy.toml"

RandomRange = t.Tuple(t.ToFloat, t.ToFloat)
num_core = 4


dummy_local_config = t.Dict({
    t.Key("agent"): t.Dict({
        t.Key("delay"): t.Dict({
            t.Key("scan-image", default=0.1): tx.Delay,
            t.Key("pull-image", default=1.0): tx.Delay,
            t.Key("push-image", default=1.0): tx.Delay,
            t.Key("purge-images", default=1.0): tx.Delay,
            t.Key("destroy-kernel", default=1.0): tx.Delay,
            t.Key("clean-kernel", default=1.0): tx.Delay,
            t.Key("create-network", default=1.0): tx.Delay,
            t.Key("destroy-network", default=1.0): tx.Delay,
            t.Key("destroy-network", default=1.0): tx.Delay,
        }).allow_extra("*"),
        t.Key("image"): t.Dict({
            t.Key("already-have", default=None): t.Null | t.List(t.String),
            t.Key("need-to-pull", default=None): t.Null | t.List(t.String),
            t.Key("missing", default=None): t.Null | t.List(t.String),
        }),
        t.Key("resource"): t.Dict({
            t.Key("cpu"): t.Dict({
                t.Key("num-core", default=num_core): t.ToInt,
            }),
            t.Key("memory"): t.Dict({
                t.Key("size", default=34359738368): t.Int,
            }),
        }),
    }).allow_extra("*"),
    t.Key("kernel-creation-ctx"): t.Dict({
        t.Key("delay"): t.Dict({
            t.Key("prepare-scratch", default=1.0): tx.Delay,
            t.Key("prepare-ssh", default=1.0): tx.Delay,
            t.Key("spawn", default=0.5): tx.Delay,
            t.Key("start-container", default=2.0): tx.Delay,
            t.Key("mount-krunner", default=1.0): tx.Delay,
        })
    }),
    t.Key("kernel"): t.Dict({
        t.Key("use-fake-code-runner", default=True): t.Bool,
        t.Key("delay"): t.Dict({
            t.Key("check-status", default=0.1): tx.Delay,
            t.Key("get-completions", default=0.1): tx.Delay,
            t.Key("get-logs", default=0.1): tx.Delay,
            t.Key("interrupt-kernel", default=0.1): tx.Delay,
            t.Key("start-service", default=0.1): tx.Delay,
            t.Key("start-model-service", default=0.1): tx.Delay,
            t.Key("shutdown-service", default=0.1): tx.Delay,
            t.Key("commit", default=5.0): tx.Delay,
            t.Key("get-service-apps", default=0.1): tx.Delay,
            t.Key("accept-file", default=0.1): tx.Delay,
            t.Key("download-file", default=0.1): tx.Delay,
            t.Key("download-single", default=0.1): tx.Delay,
            t.Key("list-files", default=0.1): tx.Delay,
        }),
    }),
}).allow_extra("*")



"""
Agent Configuration Schema
--------------------------
"""

from __future__ import annotations

import enum
import ipaddress
import logging
import os
import sys
import textwrap
from pathlib import Path
from typing import Any, Mapping, Optional, Self

from pydantic import (
    AliasChoices,
    ConfigDict,
    Field,
    FilePath,
    PrivateAttr,
    ValidationInfo,
    field_validator,
    model_validator,
)

from ai.backend.agent.utils import get_arch_name
from ai.backend.common.config import BaseConfigSchema
from ai.backend.common.configs.redis import RedisConfig
from ai.backend.common.configs.validation_context import BaseConfigValidationContext
from ai.backend.common.data.config.types import EtcdConfigData
from ai.backend.common.typed_validators import (
    AutoDirectoryPath,
    GroupID,
    HostPortPair,
    UserID,
)
from ai.backend.common.types import (
    BinarySize,
    BinarySizeField,
    ResourceGroupType,
    ServiceDiscoveryType,
)
from ai.backend.logging import BraceStyleAdapter
from ai.backend.logging.config import LoggingConfig

from ..affinity_map import AffinityPolicy
from ..stats import StatModes
from ..types import AgentBackend

log = BraceStyleAdapter(logging.getLogger(__spec__.name))


class EventLoopType(enum.StrEnum):
    ASYNCIO = "asyncio"
    UVLOOP = "uvloop"


class ContainerSandboxType(enum.StrEnum):
    DOCKER = "docker"
    JAIL = "jail"


class ScratchType(enum.StrEnum):
    HOSTDIR = "hostdir"
    HOSTFILE = "hostfile"
    MEMORY = "memory"
    K8S_NFS = "k8s-nfs"


class AgentConfigValidationContext(BaseConfigValidationContext):
    is_invoked_subcommand: bool


class SyncContainerLifecyclesConfig(BaseConfigSchema):
    enabled: bool = Field(
        default=True,
        description="Whether to enable container lifecycle synchronization",
        examples=[True, False],
    )
    interval: float = Field(
        default=10.0,
        ge=0,
        description="Synchronization interval in seconds",
        examples=[10.0, 30.0],
    )


class PyroscopeConfig(BaseConfigSchema):
    enabled: bool = Field(
        default=False,
        description="Whether to enable Pyroscope profiling",
        examples=[True, False],
    )
    app_name: Optional[str] = Field(
        default=None,
        description="Application name to use in Pyroscope",
        examples=["backendai-agent"],
        validation_alias=AliasChoices("app-name", "app_name"),
        serialization_alias="app-name",
    )
    server_addr: Optional[str] = Field(
        default=None,
        description="Address of the Pyroscope server",
        examples=["http://localhost:4040"],
        validation_alias=AliasChoices("server-addr", "server_addr"),
        serialization_alias="server-addr",
    )
    sample_rate: Optional[int] = Field(
        default=None,
        description="Sampling rate for Pyroscope profiling",
        examples=[10, 100],
        validation_alias=AliasChoices("sample-rate", "sample_rate"),
        serialization_alias="sample-rate",
    )


class OTELConfig(BaseConfigSchema):
    enabled: bool = Field(
        default=False,
        description="Whether to enable OpenTelemetry",
        examples=[True, False],
    )
    log_level: str = Field(
        default="INFO",
        description="Log level for OpenTelemetry",
        examples=["DEBUG", "INFO", "WARN", "ERROR"],
        validation_alias=AliasChoices("log-level", "log_level"),
        serialization_alias="log-level",
    )
    endpoint: str = Field(
        default="http://127.0.0.1:4317",
        description="OTLP endpoint for sending traces",
        examples=["http://127.0.0.1:4317"],
    )


class ServiceDiscoveryConfig(BaseConfigSchema):
    type: ServiceDiscoveryType = Field(
        default=ServiceDiscoveryType.REDIS,
        description="Type of service discovery to use",
        examples=[item.value for item in ServiceDiscoveryType],
    )


class CoreDumpConfig(BaseConfigSchema):
    enabled: bool = Field(
        default=False,
        description="Whether to enable core dump collection",
        examples=[True, False],
    )
    path: AutoDirectoryPath = Field(
        default=AutoDirectoryPath("./coredumps"),
        description="Directory path for storing core dumps",
        examples=["./coredumps"],
    )
    backup_count: int = Field(
        default=10,
        ge=1,
        description="Number of core dump backups to retain",
        examples=[10, 20],
        validation_alias=AliasChoices("backup-count", "backup_count"),
        serialization_alias="backup-count",
    )
    size_limit: BinarySizeField = Field(
        default=BinarySize.finite_from_str("64M"),
        description="Maximum size limit for core dumps",
        examples=["64M", "128M"],
        validation_alias=AliasChoices("size-limit", "size_limit"),
        serialization_alias="size-limit",
    )
    _core_path: Optional[Path] = PrivateAttr(default=None)

    @property
    def core_path(self) -> Path:
        """
        Get the core path for core dumps.
        If not set, it returns the default path.
        """
        if self._core_path is None:
            raise ValueError(
                "Core path is not set. Please call set_core_path() before accessing core_path."
            )
        return self._core_path

    @core_path.setter
    def core_path(self, core_path: Path) -> None:
        """
        Set the core path for core dumps.
        This is used to set the core pattern file path.
        """
        self._core_path = core_path

    model_config = ConfigDict(
        arbitrary_types_allowed=True,
    )

    @model_validator(mode="after")
    def _set_coredump_path(self, info: ValidationInfo) -> Self:
        if self.enabled:
            context = AgentConfigValidationContext.get_config_validation_context(info)
            if context is None:
                raise ValueError("context must be specified in model_validate()")
            if context.is_invoked_subcommand and not sys.platform.startswith("linux"):
                raise ValueError("Storing container coredumps is only supported in Linux.")

            core_pattern = Path("/proc/sys/kernel/core_pattern").read_text().strip()
            if core_pattern.startswith("|") or not core_pattern.startswith("/"):
                raise ValueError(
                    "/proc/sys/kernel/core_pattern must be an absolute path to enable container coredumps.",
                )

            self._core_path = Path(core_pattern).parent
        return self


class DebugConfig(BaseConfigSchema):
    enabled: bool = Field(
        default=False,
        description="Master switch for debug mode",
        examples=[True, False],
    )
    asyncio: bool = Field(
        default=False,
        description="Whether to enable asyncio debug mode",
        examples=[True, False],
    )
    kernel_runner: bool = Field(
        default=False,
        description="Whether to enable kernel runner debug mode",
        examples=[True, False],
        validation_alias=AliasChoices("kernel-runner", "kernel_runner"),
        serialization_alias="kernel-runner",
    )
    enhanced_aiomonitor_task_info: bool = Field(
        default=False,
        description="Enable enhanced task information in aiomonitor",
        examples=[True, False],
        validation_alias=AliasChoices(
            "enhanced-aiomonitor-task-info", "enhanced_aiomonitor_task_info"
        ),
        serialization_alias="enhanced-aiomonitor-task-info",
    )
    skip_container_deletion: bool = Field(
        default=False,
        description="Whether to skip container deletion for debugging",
        examples=[True, False],
        validation_alias=AliasChoices("skip-container-deletion", "skip_container_deletion"),
        serialization_alias="skip-container-deletion",
    )
    log_stats: bool = Field(
        default=False,
        description="Whether to log statistics",
        examples=[True, False],
        validation_alias=AliasChoices("log-stats", "log_stats"),
        serialization_alias="log-stats",
    )
    log_kernel_config: bool = Field(
        default=False,
        description="Whether to log kernel configuration",
        examples=[True, False],
        validation_alias=AliasChoices("log-kernel-config", "log_kernel_config"),
        serialization_alias="log-kernel-config",
    )
    log_alloc_map: bool = Field(
        default=False,
        description="Whether to log allocation map",
        examples=[True, False],
        validation_alias=AliasChoices("log-alloc-map", "log_alloc_map"),
        serialization_alias="log-alloc-map",
    )
    log_events: bool = Field(
        default=False,
        description="Whether to log all internal events",
        examples=[True, False],
        validation_alias=AliasChoices("log-events", "log_events"),
        serialization_alias="log-events",
    )
    log_heartbeats: bool = Field(
        default=False,
        description="Whether to log heartbeats",
        examples=[True, False],
        validation_alias=AliasChoices("log-heartbeats", "log_heartbeats"),
        serialization_alias="log-heartbeats",
    )
    heartbeat_interval: float = Field(
        default=20.0,
        description="Heartbeat interval in seconds",
        examples=[20.0, 30.0],
        validation_alias=AliasChoices("heartbeat-interval", "heartbeat_interval"),
        serialization_alias="heartbeat-interval",
    )
    log_docker_events: bool = Field(
        default=False,
        description="Whether to log Docker events",
        examples=[True, False],
        validation_alias=AliasChoices("log-docker-events", "log_docker_events"),
        serialization_alias="log-docker-events",
    )
    coredump: CoreDumpConfig = Field(
        default_factory=CoreDumpConfig,
        description="Core dump configuration",
    )

    @field_validator("enabled", mode="before")
    @classmethod
    def _set_enabled(cls, v: Any, info: ValidationInfo) -> Any:
        context = AgentConfigValidationContext.get_config_validation_context(info)
        if context is None:
            # Likely in tests, command line args do not need to be set.
            return v
        return context.debug


class AgentConfig(BaseConfigSchema):
    backend: AgentBackend = Field(
        description=textwrap.dedent("""
        Backend type for the agent.
        This determines how the agent interacts with the underlying infrastructure.
        Available options are:
        - `docker`: Uses Docker as the backend.
        - `kubernetes`: Uses Kubernetes as the backend.
        - `dummy`: A dummy backend for testing purposes.
        """),
        examples=[item.value for item in AgentBackend],
        validation_alias=AliasChoices("backend", "mode"),
        serialization_alias="backend",
    )
    rpc_listen_addr: HostPortPair = Field(
        default=HostPortPair(host="0.0.0.0", port=6001),
        description="RPC listen address and port",
        examples=[{"host": "", "port": 6001}],
        validation_alias=AliasChoices("rpc-listen-addr", "rpc_listen_addr"),
        serialization_alias="rpc-listen-addr",
    )
    service_addr: HostPortPair = Field(
        default=HostPortPair(host="0.0.0.0", port=6003),
        description="Service address and port",
        validation_alias=AliasChoices("service-addr", "internal-addr", "service_addr"),
        serialization_alias="service-addr",
    )
    announce_addr: HostPortPair = Field(
        default=HostPortPair(host="host.docker.internal", port=6003),
        description="Announce address and port",
        validation_alias=AliasChoices("announce-addr", "announce-internal-addr", "announce_addr"),
        serialization_alias="announce-addr",
    )
    ssl_enabled: bool = Field(
        default=False,
        description="Whether to enable SSL",
        examples=[True, False],
        validation_alias=AliasChoices("ssl-enabled", "ssl_enabled"),
        serialization_alias="ssl-enabled",
    )
    ssl_cert: Optional[FilePath] = Field(
        default=None,
        description="Path to SSL certificate file",
        examples=["/path/to/cert.pem"],
        validation_alias=AliasChoices("ssl-cert", "ssl_cert"),
        serialization_alias="ssl-cert",
    )
    ssl_key: Optional[FilePath] = Field(
        default=None,
        description="Path to SSL private key file",
        examples=["/path/to/key.pem"],
        validation_alias=AliasChoices("ssl-key", "ssl_key"),
        serialization_alias="ssl-key",
    )
    advertised_rpc_addr: Optional[HostPortPair] = Field(
        default=None,
        description="Advertised RPC address and port",
        examples=[{"host": "192.168.1.100", "port": 6001}],
        validation_alias=AliasChoices("advertised-rpc-addr", "advertised_rpc_addr"),
        serialization_alias="advertised-rpc-addr",
    )
    rpc_auth_manager_public_key: Optional[FilePath] = Field(
        default=None,
        description="Path to RPC auth manager public key",
        examples=["/path/to/public.key"],
        validation_alias=AliasChoices("rpc-auth-manager-public-key", "rpc_auth_manager_public_key"),
        serialization_alias="rpc-auth-manager-public-key",
    )
    rpc_auth_agent_keypair: Optional[FilePath] = Field(
        default=None,
        description="Path to RPC auth agent keypair",
        examples=["/path/to/keypair.key"],
        validation_alias=AliasChoices("rpc-auth-agent-keypair", "rpc_auth_agent_keypair"),
        serialization_alias="rpc-auth-agent-keypair",
    )
    agent_sock_port: int = Field(
        default=6007,
        ge=1024,
        le=65535,
        description="Agent socket port",
        examples=[6007],
        validation_alias=AliasChoices("agent-sock-port", "agent_sock_port"),
        serialization_alias="agent-sock-port",
    )
    id: Optional[str] = Field(
        default=None,
        description="Agent ID",
        examples=["agent-001"],
    )
    ipc_base_path: AutoDirectoryPath = Field(
        default=AutoDirectoryPath("/tmp/backend.ai/ipc"),
        description="Base path for IPC",
        examples=["/tmp/backend.ai/ipc"],
        validation_alias=AliasChoices("ipc-base-path", "ipc_base_path"),
        serialization_alias="ipc-base-path",
    )
    var_base_path: AutoDirectoryPath = Field(
        default=AutoDirectoryPath("./var/lib/backend.ai"),
        description="Base path for variable data",
        examples=["./var/lib/backend.ai"],
        validation_alias=AliasChoices("var-base-path", "var_base_path"),
        serialization_alias="var-base-path",
    )
    mount_path: Optional[AutoDirectoryPath] = Field(
        default=None,
        description="Mount path for containers",
        examples=["/mnt/backend.ai"],
        validation_alias=AliasChoices("mount-path", "mount_path"),
        serialization_alias="mount-path",
    )
    cohabiting_storage_proxy: bool = Field(
        default=True,
        description="Whether to enable cohabiting storage proxy",
        examples=[True, False],
        validation_alias=AliasChoices("cohabiting-storage-proxy", "cohabiting_storage_proxy"),
        serialization_alias="cohabiting-storage-proxy",
    )
    public_host: Optional[str] = Field(
        default=None,
        description="Public host address",
        examples=["backend.ai"],
        validation_alias=AliasChoices("public-host", "public_host"),
        serialization_alias="public-host",
    )
    region: Optional[str] = Field(
        default=None,
        description="Region name",
        examples=["us-east-1"],
    )
    instance_type: Optional[str] = Field(
        default=None,
        description="Instance type",
        examples=["m5.large"],
        validation_alias=AliasChoices("instance-type", "instance_type"),
        serialization_alias="instance-type",
    )
    scaling_group: str = Field(
        default="default",
        description="Scaling group name",
        examples=["default", "gpu-group"],
        validation_alias=AliasChoices("scaling-group", "scaling_group"),
        serialization_alias="scaling-group",
    )
    scaling_group_type: ResourceGroupType = Field(
        default=ResourceGroupType.COMPUTE,
        description="Scaling group type",
        examples=[item.value for item in ResourceGroupType],
        validation_alias=AliasChoices("scaling-group-type", "scaling_group_type"),
        serialization_alias="scaling-group-type",
    )
    pid_file: Path = Field(
        default=Path(os.devnull),
        description="Path to PID file",
        examples=["/var/run/agent.pid"],
        validation_alias=AliasChoices("pid-file", "pid_file"),
        serialization_alias="pid-file",
    )
    event_loop: EventLoopType = Field(
        default=EventLoopType.ASYNCIO,
        description="Event loop type",
        examples=[item.value for item in EventLoopType],
        validation_alias=AliasChoices("event-loop", "event_loop"),
        serialization_alias="event-loop",
    )
    skip_manager_detection: bool = Field(
        default=False,
        description="Whether to skip manager detection",
        examples=[True, False],
        validation_alias=AliasChoices("skip-manager-detection", "skip_manager_detection"),
        serialization_alias="skip-manager-detection",
    )
    aiomonitor_termui_port: int = Field(
        default=38200,
        ge=1,
        le=65535,
        description="Port for aiomonitor terminal UI",
        examples=[38200],
        validation_alias=AliasChoices(
            "aiomonitor-termui-port", "aiomonitor-port", "aiomonitor_termui_port"
        ),
        serialization_alias="aiomonitor-termui-port",
    )
    aiomonitor_webui_port: int = Field(
        default=39200,
        ge=1,
        le=65535,
        description="Port for aiomonitor web UI",
        examples=[39200],
        validation_alias=AliasChoices("aiomonitor-webui-port", "aiomonitor_webui_port"),
        serialization_alias="aiomonitor-webui-port",
    )
    metadata_server_bind_host: str = Field(
        default="0.0.0.0",
        description="Metadata server bind host",
        examples=["0.0.0.0"],
        validation_alias=AliasChoices("metadata-server-bind-host", "metadata_server_bind_host"),
        serialization_alias="metadata-server-bind-host",
    )
    metadata_server_port: int = Field(
        default=40128,
        ge=1,
        le=65535,
        description="Metadata server port",
        examples=[40128],
        validation_alias=AliasChoices("metadata-server-port", "metadata_server_port"),
        serialization_alias="metadata-server-port",
    )
    allow_compute_plugins: Optional[set[str]] = Field(
        default=None,
        description="Allowed compute plugins",
        examples=[{"ai.backend.activator.agent", "ai.backend.accelerator.cuda_open"}],
        validation_alias=AliasChoices("allow-compute-plugins", "allow_compute_plugins"),
        serialization_alias="allow-compute-plugins",
    )
    block_compute_plugins: Optional[set[str]] = Field(
        default=None,
        description="Blocked compute plugins",
        examples=[{"ai.backend.accelerator.mock"}],
        validation_alias=AliasChoices("block-compute-plugins", "block_compute_plugins"),
        serialization_alias="block-compute-plugins",
    )
    allow_network_plugins: Optional[set[str]] = Field(
        default=None,
        description="Allowed network plugins",
        examples=[{"ai.backend.manager.network.overlay"}],
        validation_alias=AliasChoices("allow-network-plugins", "allow_network_plugins"),
        serialization_alias="allow-network-plugins",
    )
    block_network_plugins: Optional[set[str]] = Field(
        default=None,
        description="Blocked network plugins",
        examples=[{"ai.backend.manager.network.overlay"}],
        validation_alias=AliasChoices("block-network-plugins", "block_network_plugins"),
        serialization_alias="block-network-plugins",
    )
    image_commit_path: AutoDirectoryPath = Field(
        default=AutoDirectoryPath("./tmp/backend.ai/commit"),
        description="Path for image commit",
        examples=["./tmp/backend.ai/commit"],
        validation_alias=AliasChoices("image-commit-path", "image_commit_path"),
        serialization_alias="image-commit-path",
    )
    abuse_report_path: Optional[Path] = Field(
        default=None,
        description="Path for abuse reports",
        examples=["/var/log/backend.ai/abuse"],
        validation_alias=AliasChoices("abuse-report-path", "abuse_report_path"),
        serialization_alias="abuse-report-path",
    )
    force_terminate_abusing_containers: bool = Field(
        default=False,
        description="Whether to force terminate abusing containers",
        examples=[True, False],
        validation_alias=AliasChoices(
            "force-terminate-abusing-containers", "force_terminate_abusing_containers"
        ),
        serialization_alias="force-terminate-abusing-containers",
    )
    kernel_creation_concurrency: int = Field(
        default=4,
        ge=1,
        le=32,
        description="Kernel creation concurrency",
        examples=[4, 8],
        validation_alias=AliasChoices("kernel-creation-concurrency", "kernel_creation_concurrency"),
        serialization_alias="kernel-creation-concurrency",
    )
    use_experimental_redis_event_dispatcher: bool = Field(
        default=False,
        description="Whether to use experimental Redis event dispatcher",
        examples=[True, False],
        validation_alias=AliasChoices(
            "use-experimental-redis-event-dispatcher", "use_experimental_redis_event_dispatcher"
        ),
        serialization_alias="use-experimental-redis-event-dispatcher",
    )
    sync_container_lifecycles: SyncContainerLifecyclesConfig = Field(
        default_factory=SyncContainerLifecyclesConfig,
        description="Container lifecycle synchronization config",
        validation_alias=AliasChoices("sync-container-lifecycles", "sync_container_lifecycles"),
        serialization_alias="sync-container-lifecycles",
    )
    docker_mode: Optional[str] = Field(
        default=None,
        description="Docker mode detected based on kernel version (linuxkit/native)",
        examples=["linuxkit", "native"],
        validation_alias=AliasChoices("docker-mode", "docker_mode"),
        serialization_alias="docker-mode",
    )
    mount_path_uid_gid: Optional[str] = Field(
        default=None,
        description="Owner uid:gid of the mount directory",
        examples=["root:root", "bai:bai"],
        validation_alias=AliasChoices("mount-path-uid-gid", "mount_path_uid_gid"),
        serialization_alias="mount-path-uid-gid",
    )

    def real_mount_path(self, directory_path: str) -> Path:
        if self.mount_path is None:
            return Path("./", directory_path)
        return Path(self.mount_path, directory_path)

    model_config = ConfigDict(
        extra="allow",
    )

    @field_validator("rpc_listen_addr", mode="after")
    @classmethod
    def _validate_rpc_listen_addr(cls, rpc_listen_addr: HostPortPair) -> HostPortPair:
        try:
            rpc_host = ipaddress.ip_address(rpc_listen_addr.host)
        except ValueError:
            return rpc_listen_addr
        if rpc_host.is_link_local:
            raise ValueError("Cannot use link-local IP address as the RPC listening host.")
        return rpc_listen_addr


class ContainerConfig(BaseConfigSchema):
    kernel_uid: UserID = Field(
        default=UserID(-1),
        description="Kernel user ID",
        examples=[1000, -1],
        validation_alias=AliasChoices("kernel-uid", "kernel_uid"),
        serialization_alias="kernel-uid",
    )
    kernel_gid: GroupID = Field(
        default=GroupID(-1),
        description="Kernel group ID",
        examples=[1000, -1],
        validation_alias=AliasChoices("kernel-gid", "kernel_gid"),
        serialization_alias="kernel-gid",
    )
    bind_host: str = Field(
        default="",
        description="Bind host for containers",
        examples=["0.0.0.0", ""],
        validation_alias=AliasChoices("bind-host", "bind_host", "kernel-host"),
        serialization_alias="bind-host",
    )
    advertised_host: Optional[str] = Field(
        default=None,
        description="Advertised host for containers",
        examples=["192.168.1.100"],
        validation_alias=AliasChoices("advertised-host", "advertised_host"),
        serialization_alias="advertised-host",
    )
    port_range: tuple[int, int] = Field(
        default=(30000, 31000),
        description="Port range for containers",
        examples=[(30000, 31000)],
        validation_alias=AliasChoices("port-range", "port_range"),
        serialization_alias="port-range",
    )
    stats_type: Optional[StatModes] = Field(
        default=StatModes.DOCKER,
        description="Statistics type",
        examples=[item.value for item in StatModes],
        validation_alias=AliasChoices("stats-type", "stats_type"),
        serialization_alias="stats-type",
    )
    sandbox_type: ContainerSandboxType = Field(
        default=ContainerSandboxType.DOCKER,
        description="Sandbox type",
        examples=[item.value for item in ContainerSandboxType],
        validation_alias=AliasChoices("sandbox-type", "sandbox_type"),
        serialization_alias="sandbox-type",
    )
    jail_args: list[str] = Field(
        default_factory=list,
        description="Jail arguments",
        examples=[["--mount", "/tmp"]],
        validation_alias=AliasChoices("jail-args", "jail_args"),
        serialization_alias="jail-args",
    )
    scratch_type: ScratchType = Field(
        description="Scratch type",
        examples=[item.value for item in ScratchType],
        validation_alias=AliasChoices("scratch-type", "scratch_type"),
        serialization_alias="scratch-type",
    )
    scratch_root: AutoDirectoryPath = Field(
        default=AutoDirectoryPath("./scratches"),
        description="Scratch root directory",
        examples=["./scratches"],
        validation_alias=AliasChoices("scratch-root", "scratch_root"),
        serialization_alias="scratch-root",
    )
    scratch_size: BinarySizeField = Field(
        default=BinarySize.finite_from_str("0"),
        description="Scratch size",
        examples=["1G", "500M"],
        validation_alias=AliasChoices("scratch-size", "scratch_size"),
        serialization_alias="scratch-size",
    )

    scratch_nfs_address: Optional[str] = Field(
        default=None,
        description="Scratch NFS address",
        examples=["192.168.1.100:/export"],
        validation_alias=AliasChoices("scratch-nfs-address", "scratch_nfs_address"),
        serialization_alias="scratch-nfs-address",
    )
    scratch_nfs_options: Optional[str] = Field(
        default=None,
        description="Scratch NFS options",
        examples=["rw,sync"],
        validation_alias=AliasChoices("scratch-nfs-options", "scratch_nfs_options"),
        serialization_alias="scratch-nfs-options",
    )
    alternative_bridge: Optional[str] = Field(
        default=None,
        description="Alternative bridge network",
        examples=["br-backend"],
        validation_alias=AliasChoices("alternative-bridge", "alternative_bridge"),
        serialization_alias="alternative-bridge",
    )
    krunner_volumes: Optional[Mapping[str, str]] = Field(
        default=None,
        description=textwrap.dedent("""
        KRunner volumes configuration, mapping container names to host paths.
        This is used to specify volumes that should be mounted into containers
        when using the KRunner backend.
        This fields is filled by the agent at runtime based on the
        `krunner_volumes` configuration in the agent's environment.
        It is not intended to be set in the configuration file.
        """),
        examples=[],
        validation_alias=AliasChoices("krunner-volumes", "krunner_volumes"),
        serialization_alias="krunner-volumes",
    )
    swarm_enabled: bool = Field(
        default=False,
        description=textwrap.dedent("""
        Whether to enable Docker Swarm mode.
        This allows the agent to manage containers in a Docker Swarm cluster.
        When enabled, the agent will use Docker Swarm APIs to manage containers,
        networks, and services.
        This field is only used when backend is set to 'docker'.
        """),
        examples=[True, False],
        validation_alias=AliasChoices("swarm-enabled", "swarm_enabled"),
        serialization_alias="swarm-enabled",
    )

    model_config = ConfigDict(
        extra="allow",
        arbitrary_types_allowed=True,
    )

    @field_validator("port_range", mode="before")
    @classmethod
    def _parse_port_range(cls, v: Any) -> tuple[int, int]:
        if isinstance(v, (list, tuple)) and len(v) == 2:
            return (int(v[0]), int(v[1]))
        raise ValueError("port_range must be a tuple of two integers")

    @field_validator("sandbox_type", mode="after")
    @classmethod
    def _validate_sandbox_type(cls, sandbox_type: ContainerSandboxType) -> ContainerSandboxType:
        # FIXME: Remove this after ARM64 support lands on Jail
        if sandbox_type == ContainerSandboxType.JAIL:
            current_arch = get_arch_name()
            if current_arch != "x86_64":
                raise ValueError(f"Jail sandbox is not supported on architecture {current_arch}")
        return sandbox_type

    @field_validator("stats_type", mode="after")
    @classmethod
    def _validate_stats_type(cls, stats_type: StatModes) -> StatModes:
        if stats_type == StatModes.CGROUP:
            if os.getuid() != 0:
                raise ValueError(
                    "Cannot use cgroup statistics collection mode unless the agent runs as root."
                )
        return stats_type

    @field_validator("scratch_type", mode="after")
    @classmethod
    def _validate_scratch_type(cls, scratch_type: ScratchType) -> ScratchType:
        if scratch_type == ScratchType.HOSTFILE:
            if os.getuid() != 0:
                raise ValueError("Cannot use hostfile scratch type unless the agent runs as root.")
        return scratch_type


class ResourceConfig(BaseConfigSchema):
    reserved_cpu: int = Field(
        default=1,
        description="The number of CPU cores reserved for the operating system and the agent service.",
        examples=[1, 2],
        validation_alias=AliasChoices("reserved-cpu", "reserved_cpu"),
        serialization_alias="reserved-cpu",
    )
    reserved_mem: BinarySizeField = Field(
        default=BinarySize.finite_from_str("1G"),
        description=(
            "The memory space reserved for the operating system and the agent service. "
            "It is subtracted from the reported main memory size and not available for user workload allocation. "
            "Depending on the memory-align-size option and system configuration, "
            "this may not be the exact value but have slightly less or more values within the memory-align-size."
        ),
        examples=["1G", "2G"],
        validation_alias=AliasChoices("reserved-mem", "reserved_mem"),
        serialization_alias="reserved-mem",
    )
    reserved_disk: BinarySizeField = Field(
        default=BinarySize.finite_from_str("8G"),
        description=(
            "The disk space reserved for the operating system and the agent service. "
            "Currently this value is unused. "
            "In future releases, it may be used to preserve the minimum disk space "
            "from the scratch disk allocation via loopback files."
        ),
        examples=["8G", "16G"],
        validation_alias=AliasChoices("reserved-disk", "reserved_disk"),
        serialization_alias="reserved-disk",
    )
    memory_align_size: BinarySizeField = Field(
        default=BinarySize.finite_from_str("16M"),
        description=(
            "The alignment of the reported main memory size to absorb tiny deviations "
            "from per-node firwmare/hardware settings. "
            "Recommended to be multiple of the page/hugepage size (e.g., 2 MiB)."
        ),
        examples=["2M", "32M"],
        validation_alias=AliasChoices("memory-align-size", "memory_align_size"),
        serialization_alias="memory-align-size",
    )
    allocation_order: list[str] = Field(
        default=["cuda", "rocm", "tpu", "cpu", "mem"],
        description="Resource allocation order",
        examples=[["cuda", "rocm", "tpu", "cpu", "mem"]],
        validation_alias=AliasChoices("allocation-order", "allocation_order"),
        serialization_alias="allocation-order",
    )
    affinity_policy: AffinityPolicy = Field(
        default=AffinityPolicy.INTERLEAVED,
        description="Affinity policy",
        examples=[item.name for item in AffinityPolicy],
        validation_alias=AliasChoices("affinity-policy", "affinity_policy"),
        serialization_alias="affinity-policy",
    )

    model_config = ConfigDict(
        extra="allow",
        arbitrary_types_allowed=True,
    )

    @field_validator("affinity_policy", mode="before")
    @classmethod
    def _parse_affinity_policy(cls, v: Any) -> AffinityPolicy:
        if isinstance(v, str):
            try:
                return AffinityPolicy[v.upper()]
            except KeyError:
                raise ValueError(f"Invalid affinity policy: {v}")
        return v


class EtcdConfig(BaseConfigSchema):
    namespace: str = Field(
        description="Etcd namespace",
        examples=["local", "backend"],
    )
    addr: HostPortPair | list[HostPortPair] = Field(
        description="Etcd address and port",
        examples=[
            {"host": "127.0.0.1", "port": 2379},  # single endpoint
            [
                {"host": "127.0.0.4", "port": 2379},
                {"host": "127.0.0.5", "port": 2379},
            ],  # multiple endpoints
        ],
    )
    user: Optional[str] = Field(
        default=None,
        description="Etcd username",
        examples=["backend"],
    )
    password: Optional[str] = Field(
        default=None,
        description="Etcd password",
        examples=["PASSWORD"],
    )

    def to_dataclass(self) -> EtcdConfigData:
        return EtcdConfigData(
            namespace=self.namespace,
            addrs=self.addr if isinstance(self.addr, list) else [self.addr],
            user=self.user,
            password=self.password,
        )


class ContainerLogsConfig(BaseConfigSchema):
    max_length: BinarySizeField = Field(
        default=BinarySize.finite_from_str("10M"),
        description="Maximum length of container logs",
        examples=["10M", "50M"],
        validation_alias=AliasChoices("max-length", "max_length"),
        serialization_alias="max-length",
    )
    chunk_size: BinarySizeField = Field(
        default=BinarySize.finite_from_str("64K"),
        description="Chunk size for container logs",
        examples=["64K", "128K"],
        validation_alias=AliasChoices("chunk-size", "chunk_size"),
        serialization_alias="chunk-size",
    )

    model_config = ConfigDict(
        arbitrary_types_allowed=True,
    )


class APIConfig(BaseConfigSchema):
    pull_timeout: Optional[float] = Field(
        default=7200.0,  # 2 hours
        ge=0,
        description="Image pull timeout in seconds",
        examples=[7200.0, 3600.0],
        validation_alias=AliasChoices("pull-timeout", "pull_timeout"),
        serialization_alias="pull-timeout",
    )
    commit_timeout: Optional[float] = Field(
        default=None,
        ge=0,
        description="Image commit timeout in seconds",
        examples=[7200.0, 3600.0],
        validation_alias=AliasChoices("commit-timeout", "commit_timeout"),
        serialization_alias="commit-timeout",
    )
    push_timeout: Optional[float] = Field(
        default=None,
        ge=0,
        description="Image push timeout in seconds",
        examples=[7200.0, 3600.0],
        validation_alias=AliasChoices("push-timeout", "push_timeout"),
        serialization_alias="push-timeout",
    )


class KernelLifecyclesConfig(BaseConfigSchema):
    init_polling_attempt: int = Field(
        default=10,
        description="Number of init polling attempts",
        examples=[10, 20],
        validation_alias=AliasChoices("init-polling-attempt", "init_polling_attempt"),
        serialization_alias="init-polling-attempt",
    )
    init_polling_timeout_sec: float = Field(
        default=60.0,
        description="Init polling timeout in seconds",
        examples=[60.0, 120.0],
        validation_alias=AliasChoices("init-polling-timeout-sec", "init_polling_timeout_sec"),
        serialization_alias="init-polling-timeout-sec",
    )
    init_timeout_sec: float = Field(
        default=60.0,
        description="Init timeout in seconds",
        examples=[60.0, 120.0],
        validation_alias=AliasChoices("init-timeout-sec", "init_timeout_sec"),
        serialization_alias="init-timeout-sec",
    )


class DockerExtraConfig(BaseConfigSchema):
    """
    For checking additional Docker configurations
    """

    swarm_enabled: bool = Field(
        default=False,
        description="Whether Docker Swarm is enabled",
        examples=[True, False],
        validation_alias=AliasChoices("swarm-enabled", "swarm_enabled"),
        serialization_alias="swarm-enabled",
    )


class AgentUnifiedConfig(BaseConfigSchema):
    # Local config
    agent: AgentConfig = Field(
        description="Agent configuration",
    )
    container: ContainerConfig = Field(
        description="Container configuration",
    )
    pyroscope: PyroscopeConfig = Field(
        default_factory=PyroscopeConfig,
        description="Pyroscope configuration",
    )
    logging: LoggingConfig = Field(
        default_factory=LoggingConfig,
        description="Logging configuration",
    )
    resource: ResourceConfig = Field(
        description="Resource configuration",
    )
    otel: OTELConfig = Field(
        default_factory=OTELConfig,
        description="OpenTelemetry configuration",
    )
    service_discovery: ServiceDiscoveryConfig = Field(
        default_factory=ServiceDiscoveryConfig,
        description="Service discovery configuration",
        validation_alias=AliasChoices("service-discovery", "service_discovery"),
        serialization_alias="service-discovery",
    )
    debug: DebugConfig = Field(
        default_factory=DebugConfig,
        description="Debug configuration",
    )
    etcd: EtcdConfig = Field(
        description="Etcd configuration",
    )

    # Etcd config
    container_logs: ContainerLogsConfig = Field(
        default_factory=ContainerLogsConfig,
        description="Container logs configuration",
        validation_alias=AliasChoices("container-logs", "container_logs"),
        serialization_alias="container-logs",
    )
    api: APIConfig = Field(
        default_factory=APIConfig,
        description="API configuration",
    )
    kernel_lifecycles: KernelLifecyclesConfig = Field(
        default_factory=KernelLifecyclesConfig,
        description="Kernel lifecycles configuration",
        validation_alias=AliasChoices("kernel-lifecycles", "kernel_lifecycles"),
        serialization_alias="kernel-lifecycles",
    )
    plugins: Any = Field(
        default_factory=lambda: {},
        description=textwrap.dedent("""
        Plugins configuration.
        This field is injected at runtime based on etcd configuration.
        It is not intended to be set in the configuration file.
        """),
    )
    redis: Optional[RedisConfig] = Field(
        default=None,
        description=textwrap.dedent("""
        Redis configuration.
        This field is injected at runtime based on etcd configuration.
        It is not intended to be set in the other way.
        """),
    )

    # TODO: Remove me after changing config injection logic
    model_config = ConfigDict(
        extra="allow",
    )

    @property
    def agent_backend(self) -> AgentBackend:
        return self.agent.backend

    @model_validator(mode="after")
    def _validate_kubernetes_config(self) -> Self:
        if self.agent_backend == AgentBackend.KUBERNETES:
            if self.container.scratch_type == ScratchType.K8S_NFS and (
                self.container.scratch_nfs_address is None
                or self.container.scratch_nfs_options is None
            ):
                raise ValueError(
                    "scratch-nfs-address and scratch-nfs-options are required for k8s-nfs"
                )
        return self

    @model_validator(mode="after")
    def _validate_docker_config(self) -> Self:
        if self.agent_backend == AgentBackend.DOCKER:
            DockerExtraConfig.model_validate(self.container.model_dump())
        return self
